import numpy as np

def smooth(x,window_len=10,window='hanning'):
    """smooth the data using a window with requested size.
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    Parameters
    ----------
    x : array of floats
        the input signal 
    window_len : int
        (Optional) Default=10. The dimension of the smoothing window
    window : str
        the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
        flat window will produce a moving average smoothing.
    
    Returns
    -------
    array of floats
        The smoothed signal
        
    Examples
    --------
    >>> t=linspace(-2,2,0.1)
    >>> x=sin(t)+randn(len(t))*0.1
    >>> y=smooth(x)
    
    See Also
    -------- 
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter
 
    Todos
    -----
    The window parameter could be the window itself if an array instead of a string
	Source: http://www.scipy.org/Cookbook/SignalSmooth		2009-03-13 
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")


    if window_len<3:
        return x


    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is one of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")


    #s=numpy.r_[2*x[0]-x[window_len:1:-1],x,2*x[-1]-x[-1:-window_len:-1]]
    s=np.r_[2*np.median(x[0:int(window_len/5)])-x[window_len:1:-1],x,2*np.median(x[-int(window_len/5):])-x[-1:-window_len:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w=ones(window_len,'d')
    else:
        w=eval('np.'+window+'(window_len)')

    y=np.convolve(w/w.sum(),s,mode='same')
    return y[window_len-1:-window_len+1]

def medfilt(x, window_len):
    """Apply a length-k median filter to a 1D array x.
    Boundaries are extended by repeating endpoints.
    """
    assert x.ndim == 1, "Input must be one-dimensional."
    if window_len % 2 == 0:
        print(("Median filter length ("+str(window_len)+") must be odd. Adding 1."))
        window_len += 1
    k2 = (window_len - 1) // 2
    s=np.r_[2*np.median(x[0:int(window_len/5)])-x[window_len:1:-1],x,2*np.median(x[-int(window_len/5):])-x[-1:-window_len:-1]]
    y = np.zeros ((len (s), window_len), dtype=s.dtype)
    y[:,k2] = s
    for i in range (k2):
        j = k2 - i
        y[j:,i] = s[:-j]
        y[:j,i] = s[0]
        y[:-j,-(i+1)] = s[j:]
        y[-j:,-(i+1)] = s[-1]
    return np.median(y[window_len-1:-window_len+1], axis=1)


# This is a special __init__.py required to for namespace packages.
# There should be no other code in this module.
try:
    from pkg_resources import declare_namespace
    declare_namespace(__name__)
except ImportError:
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)




import os 
import pandas as pd
from sqlalchemy import *
import datetime

def jwst_log(final):
    '''Logging for Website;Used for tracking usage;Only storing instrument data'''
    try:
        logs = os.environ.get('pandexo_logs')
        engine = create_engine('sqlite:///' + os.path.join(logs))
    except: 
        pass 
    instrument = final['pandeia_input']['configuration']['instrument']['instrument']
    mode = final['pandeia_input']['configuration']['instrument']['mode']
    ff = final['pandeia_input']['configuration']['instrument']['filter']
    if ff == None: ff = 'None'
    aperture = final['pandeia_input']['configuration']['instrument']['aperture']
    disperser = final['pandeia_input']['configuration']['instrument']['disperser']
    subarray = final['pandeia_input']['configuration']['detector']['subarray']
    if subarray == None: subarray = "None"
    calc_type= final['pandexo_input']['planet']['type']
    time =  datetime.datetime.now().strftime("%Y-%m-%d %H:%M")

    pddb = pd.DataFrame({'instrument':[instrument], 'calc_type':[calc_type],
    	'mode':[mode],'ff':[ff], 'aperture':[aperture],'disperser':[disperser],'subarray':[subarray]},index=[time])
    #print(pddb)
    pddb.to_sql('jwst_cycle1',engine,if_exists='append')

    return


def hst_log(final):
    '''Logging for Website;Used for tracking usage;Used for tracking usage;Only storing instrument data'''
    try:
        logs = os.environ.get('pandexo_logs')
        engine = create_engine('sqlite:///' + os.path.join(logs))
    except: 
        pass 
    instrument = final['pandeia_input']['configuration']['instrument']['instrument']
    subarray = final['pandeia_input']['configuration']['detector']['subarray']
    nsamp = final['pandeia_input']['configuration']['detector']['nsamp']
    samp_seq = final['pandeia_input']['configuration']['detector']['samp_seq']
    disperser = final['pandeia_input']["configuration"]['instrument']['disperser']
    scanDirection = final['pandeia_input']["strategy"]["scanDirection"]
    schedulability = final['pandeia_input']["strategy"]["schedulability"]

    calc_type= final['pandexo_input']['planet']['type']


    time =  datetime.datetime.now().strftime("%Y-%m-%d %H:%M")

    pddb = pd.DataFrame({'instrument':[instrument], 'calc_type':[calc_type],
    	'nsamp':[nsamp], 'samp_seq':[samp_seq],'disperser':[disperser],'subarray':[subarray],'scanDirection':[scanDirection],
    	'schedulability':[schedulability]},index=[time])
    #print(pddb)
    pddb.to_sql('hst_cycle26',engine,if_exists='append')
    return



import itertools
import os
import re
import requests
import urllib


def build_target_url(target_name):
    '''Build restful api url based on target name.
    Parameters
        ----------
        target_name : string
            The name of the target transit.
        Returns
        -------
        target_url : string
    '''
    # Encode the target name string.
    encode_target_name = urllib.parse.quote(target_name, encoding='utf-8')
    target_url = "https://exo.mast.stsci.edu/api/v0.1/exoplanets/{}/properties/".format(encode_target_name)

    return target_url

def get_canonical_name(target_name):
    '''Get ExoMAST prefered name for exoplanet.
        Parameters
        ----------
        target_name : string
            The name of the target transit.
        Returns
        -------
        canonical_name : string
    '''

    target_url = "https://exo.mast.stsci.edu/api/v0.1/exoplanets/identifiers/"

    # Create params dict for url parsing. Easier than trying to format yourself.
    params = {"name":target_name}

    r = requests.get(target_url, params=params)
    planetnames = r.json()
    canonical_name = planetnames['canonicalName']

    return canonical_name

def get_target_data(target_name):
    """
    Send request to exomast restful api for target information.
    Parameters
    ----------
    target_name : string
        The name of the target transit
    Returns
    -------
    target_data: json:
        json object with target data.
    """

    canonical_name = get_canonical_name(target_name)

    target_url = build_target_url(canonical_name)

    r = requests.get(target_url)

    if r.status_code == 200:
        target_data = r.json()
    else:
        raise Exception('Whoops, no data for this target!')

    # Some targets have multiple catalogs
    # nexsci is the first choice.
    if len(target_data) > 1:
        # Get catalog names from exomast and make then the keys of a dictionary
        # and the values are its position in the json object.
        catalog_dict = {data['catalog_name']: index for index, data in enumerate(target_data)}

        # Parse based on catalog accuracy.
        if 'nexsci' in list(catalog_dict.keys()):
            target_data = target_data[catalog_dict['nexsci']]
        elif 'exoplanets.org' in list(catalog_dict.keys()):
            target_data = target_data[catalog_dict['exoplanets.org']]
        else:
            target_data = target_data[0]
    else:
        target_data = target_data[0]

    # Strip spaces and non numeric or alphabetic characters and combine.
    url = 'https://exo.mast.stsci.edu/exomast_planet.html?planet={}'.format(re.sub(r'\W+', '', canonical_name))

    return target_data, url



import pandas as pd 
import numpy as np 
import warnings
with warnings.catch_warnings():
	warnings.filterwarnings("ignore")
	import pysynphot.binning as astrobin 
import warnings as warn
import astropy.units as u
def binning(x, y,  dy=None, binwidth=None, r=None,newx= None, log = False, nan=False):
	"""
	This contains functionality for binning spectroscopy given an x, y and set of errors. 
	This is similar to IDL's regroup but in Python (obviously). Note that y is binned as the 
	mean(ordinates) instead of sum(ordinates), as you would want for cmputing flux in a set of 
	pixels. User can input a constant resolution, constant binwidth or provide a user defined 
	bin. The error is computed as sum(sqrt(sig1^2, sig2^2, sig3^2) )/3, for example 
	if there were 3 points to bin. 
	
	Parameters
	----------
	x : array, float 
		vector containing abcissae.
	y : array,float 
		vector containing ordinates.
	dy : array,float
		(Optional) errors on ordinates, can be float or array
	binwidth : float 
		(Optional) constant bin width in same units as x 
	r : float 
		(Optional) constant resolution to bin to. R is defined as w[1]/(w[2] - w[0]) 
		to maintain consistency with `pandeia.engine`
	newx : array, float 
		(Optional) new x axis to bin to 
	log : bool
		(Optional) computes equal bin spacing logarithmically, Default = False
	sort : bool 
		(Optional) sort into ascending order of x, default = True
	nan : bool
		(Optional) if true, this returns nan values where no points exist in a given bin
		Otherwise, all nans are dropped 

	Returns
	-------
	dict 
		bin_y : binned ordinates 
		bin_x : binned abcissae 
		bin_edge : edges of bins (always contains len(bin_x)+1 elements)
		bin_dy : error on ordinate bin 
		bin_n : number of points contained in each bin 

	Examples
	--------

	>>> from bintools import binning 

	If you want constant resolution (using output dict from PandExo): 
	
	>>> pandexo = result['FinalSpectrum']
	>>> x, y, err = pandexo['wave'], pandexo['spectrum_w_rand'], pandexo['error_w_floor']
	>>> final = binning(x, y, dy = err, r =100)
	>>> newx, newy, newerr = final['bin_x'], final['bin_y'], final['bin_dy']

	If you have a x axis that you want PandExo output to be binned to

	>>> newx = np.linspace(1,5,10)
	>>> final = binning(x, y, dy = err, newx =newx)
	>>> newx, newy, newerr = final['bin_x'], final['bin_y'], final['bin_dy']

	If you want a constant bin width and want everything to be spaced linearly

	>>> final = binning(x, y, dy = err, binwidth = 0.1)
	>>> newx, newy, newerr = final['bin_x'], final['bin_y'], final['bin_dy']

	If you want constant bin width but want everything to spaced logarithmically 

	>>> final = binning(x, y, dy = err, binwidth = 0.1, log=True)
	>>> newx, newy, newerr = final['bin_x'], final['bin_y'], final['bin_dy']
	"""
	#check x and y are same length 
	if len(x) != len(y):
		raise Exception('X and Y are not the same length')

	#check that either newx or binwidth are specified 

	if newx is None and binwidth is None and r is None:
		raise Exception('Need to either supply new x axis, resolution, or a binwidth')
	if (binwidth is None) and (log): 
		raise Exception("Cannot do logarithmic binning with out a binwidth")

	if newx is not None: 
		bin_x = newx
		bin_x, bin_y, bin_dy, bin_n = uniform_tophat_mean(bin_x,x, y, dy=dy,nan=nan)
		bin_edge = astrobin.calculate_bin_edges(bin_x)

		return {'bin_y':bin_y, 'bin_x':bin_x, 'bin_edge':bin_edge, 'bin_dy':bin_dy, 'bin_n':bin_n} 

	elif r is not None:
		bin_x = bin_wave_to_R(x, r)
		bin_x, bin_y, bin_dy, bin_n = uniform_tophat_mean(bin_x,x, y, dy=dy,nan=nan)
		bin_edge = astrobin.calculate_bin_edges(bin_x)

		return {'bin_y':bin_y, 'bin_x':bin_x, 'bin_edge':bin_edge, 'bin_dy':bin_dy, 'bin_n':bin_n} 
 

	elif binwidth is not None: 
		if (binwidth < 0) and (log):
			warn.warn(UserWarning("Negative binwidth specified. Assuming this is log10(binwidth)"))
			binwidth = 10**binwidth
		if log:
			bin_x = np.arange(np.log10(min(x)),np.log10(max(x)),np.log10(binwidth))
			bin_x = 10**bin_x
		elif not log:
			bin_x = np.arange(min(x),max(x),binwidth)
		bin_x, bin_y, bin_dy, bin_n = uniform_tophat_mean(bin_x,x, y, dy=dy,nan=nan)
		bin_edge = astrobin.calculate_bin_edges(bin_x)

		return {'bin_y':bin_y, 'bin_x':bin_x, 'bin_edge':bin_edge, 'bin_dy':bin_dy, 'bin_n':bin_n} 



def uniform_tophat_mean(newx,x, y, dy=None,nan=False):
	"""Adapted from Mike R. Line to rebin spectra

	Takes mean of groups of points in certain wave bin 

	Parameters
	----------
	newx : list of float or numpy array of float
	    New wavelength grid to rebin to 
	x : list of float or numpy array of float 
	    Old wavelength grid to get rid of 
	y : list of float or numpy array of float 
	    New rebinned y axis 

	Returns
	-------
	array of floats 
	    new wavelength grid 
	    
	Examples 
	--------
	    
	>>> from pandexo.engine.jwst import uniform_tophat_sum
	>>> oldgrid = np.linspace(1,3,100)
	>>> y = np.zeros(100)+10.0
	>>> newy = uniform_tophat_sum(np.linspace(2,3,3), oldgrid, y)
	>>> newy
	array([ 240.,  250.,  130.])
	"""
	newx = np.array(newx)
	szmod=newx.shape[0]
	delta=np.zeros(szmod)
	ynew=np.zeros(szmod)
	bin_dy =np.zeros(szmod)
	bin_n =np.zeros(szmod)

	delta[0:-1]=newx[1:]-newx[:-1]  
	delta[szmod-1]=delta[szmod-2] 
    
	for i in range(szmod-1):
		i=i+1
		loc=np.where((x >= newx[i]-0.5*delta[i-1]) & (x < newx[i]+0.5*delta[i]))
		#make sure there are values within the slice 
		if len(loc[0]) > 0:
			ynew[i]=np.mean(y[loc])
			if dy is not None: 
				bin_dy[i] = np.sqrt(np.sum(dy[loc]**2.0))/len(y[loc])
			bin_n[i] = len(y[loc])
		#if not give empty slice a nan
		elif len(loc[0]) is 0 : 
			warn.warn(UserWarning("Empty slice exists within specified new x, replacing value with nan"))
			ynew[i]=np.nan
			bin_n[i] = np.nan 

	#fill in zeroth entry
	loc=np.where((x > newx[0]-0.5*delta[0]) & (x < newx[0]+0.5*delta[0]))
	if len(loc[0]) > 0: 
		ynew[0]=np.mean(y[loc])
		bin_n[0] = len(y[loc])
		if dy is not None: 
			bin_dy[0] = np.sqrt(np.sum(dy[loc]**2.0))/len(y[loc])
	elif len(loc[0]) is 0 : 
		ynew[0]=np.nan
		bin_n[0] = np.nan
		if dy is not None:
			bin_dy[0] = np.nan 
	
	#remove nans if requested
	out = pd.DataFrame({'bin_y':ynew, 'bin_x':newx, 'bin_dy':bin_dy, 'bin_n':bin_n})
	if not nan:
		out = out.dropna()

	return out['bin_x'].values,out['bin_y'].values, out['bin_dy'].values, out['bin_n'].values

def bin_wave_to_R(w, R):
	"""Creates new wavelength axis at specified resolution

	Parameters
	----------
	w : list of float or numpy array of float
	    Wavelength axis to be rebinned 
	R : float or int 
	    Resolution to bin axis to 

	Returns
	-------
	list of float
	    New wavelength axis at specified resolution

	Examples
	--------

	>>> newwave = bin_wave_to_R(np.linspace(1,2,1000), 10)
	>>> print((len(newwave)))
	11
	"""
	wave = []
	tracker = min(w)
	i = 1 
	ind= 0
	firsttime = True
	while(tracker<max(w)):
	    if i <len(w)-1:
	        dlambda = w[i]-w[ind]
	        newR = w[i]/dlambda
	        if (newR < R) & (firsttime):
	            tracker = w[ind]
	            wave += [tracker]
	            ind += 1
	            i += 1 
	            firsttime = True
	        elif newR < R:
	            tracker = w[ind]+dlambda/2.0
	            wave +=[tracker]
	            ind = (np.abs(w-tracker)).argmin()
	            i = ind+1
	            firsttime = True
	        else:
	            firsttime = False            
	            i+=1    
	    else:
	        tracker = max(w)
	        wave += [tracker]
	return wave




import numpy as np 
import pickle
import pandas as pd
from sqlalchemy import *   
import astropy.units as u 
import astropy.constants as c
import os 
from  astropy.modeling import blackbody as bb
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore")
    import pysynphot as psyn
    
def outTrans(input) :
    """Compute out of transit spectra
    
    Computes the out of transit spectra by normalizing flux to specified 
    magnitude and convert to specified Pandeia units of milliJy and microns.  
  
    Parameters 
    ----------
    input : dict
        stellar scene which includes parameters to extract phoenix database or a 
        filename which points to a stellar spectrum 
    
    Return
    ------
    dict 
        contains wave and flux_out_trans
    """ 

    ref_wave = float(input['ref_wave'])
    mag = float(input['mag'])

    ################# USER ####################################
    if input['type'] == 'user':
        if isinstance(input['starpath'], dict):
            star = input['starpath']
        else: #if isinstance(input['starpath'], str):
            star = np.genfromtxt(input['starpath'], dtype=(float, float),
                                 names='w, f')
        #get flux 
        flux = star['f'] #star.field(input['logg'])
        #get wavelength and reference wavelength for mag normalization
        wave = star['w'] #star.field('WAVELENGTH')
        
        #sort if not in ascending order 
        sort = np.array([wave,flux]).T
        sort= sort[sort[:,0].argsort()]
        wave = sort[:,0]
        flux = sort[:,1] 
        if input['w_unit'] == 'um':
            PANDEIA_WAVEUNITS = 'um'
            
        elif input['w_unit'] == 'nm':
            PANDEIA_WAVEUNITS = 'nm'
      
        elif input['w_unit'] == 'cm' :
            PANDEIA_WAVEUNITS = 'cm'
     
        elif input['w_unit'] == 'Angs' :
            PANDEIA_WAVEUNITS = 'angstrom'

        elif input['w_unit'] == 'Hz' :
            PANDEIA_WAVEUNITS = 'Hz'

        else: 
            raise Exception('Units are not correct. Pick um, nm, cm, hz, or Angs')        

        #convert to photons/s/nm/m^2 for flux normalization based on 
        #http://www.gemini.edu/sciops/instruments/integration-time-calculators/itc-help/source-definition
        if input['f_unit'] == 'Jy':
            PANDEIA_FLUXUNITS = 'jy' 
        elif input['f_unit'] == 'FLAM' :
            PANDEIA_FLUXUNITS = 'FLAM'
        elif input['f_unit'] == 'erg/cm2/s/Hz':
            flux = flux*1e23
            PANDEIA_FLUXUNITS = 'jy' 
        else: 
            raise Exception('Units are not correct. Pick FLAM or Jy or erg/cm2/s/Hz')

        sp = psyn.ArraySpectrum(wave, flux, waveunits=PANDEIA_WAVEUNITS, fluxunits=PANDEIA_FLUXUNITS)        #Convert evrything to nanometer for converstion based on gemini.edu  
        sp.convert("nm")
        sp.convert('jy')

    ############ PHOENIX ################################################
    elif input['type'] =='phoenix':
        #make sure metal is not out of bounds
        if input['metal'] > 0.5: input['metal'] = 0.5
        sp = psyn.Icat("phoenix", input['temp'], input['metal'], input['logg'])
        sp.convert("nm")
        sp.convert("jy")
        wave = sp.wave
        flux = sp.flux
        input['w_unit'] ='nm'
        input['f_unit'] = 'jy'
        
    else: 
        raise Exception('Wrong input type for stellar spectra')
    

    ############ NORMALIZATION ################################################
    refdata = os.environ.get("pandeia_refdata")

    all_bps = {"H": 'bessell_h_004_syn.fits',
                 "J":'bessell_j_003_syn.fits' ,
                 "K": 'bessell_k_003_syn.fits'}

    if (ref_wave <= 1.3) & (ref_wave >= 1.2):
        filt = 'J'
    elif (ref_wave <= 1.7) & (ref_wave >= 1.6):
        filt = 'H'
    elif (ref_wave <= 2.3) & (ref_wave >= 2.1):
        filt = 'K'
    else:
        raise Exception('Only J H and K zeropoints are included')

    bp_path = os.path.join(refdata, "normalization", "bandpass", all_bps[filt])
    bp = psyn.FileBandpass(bp_path)

    sp.convert('angstroms')
    bp.convert('angstroms')

    rn_sp = sp.renorm(mag, 'vegamag', bp)


    rn_sp.convert("microns")
    rn_sp.convert("mjy")

    flux_out_trans = rn_sp.flux
    wave = rn_sp.wave
    return {'flux_out_trans': flux_out_trans, 'wave': wave,'phoenix':sp} 


def bothTrans(out_trans, planet,star=None) :
    """Calculates in transit flux 
    
    Takes output from `outTrans`, which is the normalized stellar flux, and 
    creates either a transit transmission spectrum, phase curve or emission spectrum. 
    Magnitude 
  
    Parameters 
    ----------
    out_trans: dict 
        includes dictionary from `outTrans` output. 
    planet: dict
        dictionary with direction to planet spectra, wavelength and flux units
    star: dict
        (Optional) dictionary within exo_input with stellar information. Only 
        used when scaling Fortney Grid spectra to get (rp/r*)^2

    Return
    ------
    dict 
        dictionary with out of transit flux, in transit flux, original model 
        and corresponding wavelengths
    """ 
    
    if planet['type'] =='user':
        if isinstance(planet['exopath'], dict):
            load_file = planet['exopath']
        else: #if isinstance(planet['exopath'], str):
            load_file = np.genfromtxt(planet['exopath'], dtype=(float, float),
                names='w, f')
        #get wavelength 
        wave_planet = load_file['w']
        #get planet flux 
        flux_planet = load_file['f']
        
        #sort if not in ascending order 
        sort = np.array([wave_planet,flux_planet]).T
        sort= sort[sort[:,0].argsort()]
        wave_planet = sort[:,0]
        flux_planet = sort[:,1]
    
    ############## IF USER SELECTS CONSTANT VALUE ##################   
    elif planet['type'] == 'constant':
        rplan = (planet['radius']*u.Unit(planet['r_unit'])).to(u.km)
        rstar = (star['radius']*u.Unit(star['r_unit'])).to(u.km)

        #constant transit depth
        if planet['f_unit'] == 'rp^2/r*^2':
            wave_planet = np.linspace(0.5,15,1000)
            planet['depth'] = float(rplan**2 / rstar**2)
            flux_planet = np.linspace(0.5,15,1000)*0 + planet['depth']
            planet['w_unit'] = 'um'

        #constant fp/f* (using out_trans from user)
        elif planet['f_unit'] == 'fp/f*':
            planet['w_unit'] = 'um'
            wave_planet = out_trans['wave'][(out_trans['wave']>0.5) & (out_trans['wave']<15)]
            flux_star = (out_trans['phoenix'].flux*(u.Jy)).to(u.mJy)[(out_trans['wave']>0.5) & (out_trans['wave']<15)]
            #MAKING SURE TO ADD IN SUPID PI FOR PER STERADIAN!!!!
            flux_planet = (bb.blackbody_nu(wave_planet*u.micron, planet['temp']*u.K)*np.pi*u.sr).to(u.mJy)
            # ( bb planet / pheonix sed ) * (rp/r*)^2
            flux_planet = np.array((flux_planet/flux_star) * (rplan/rstar)**2.0)

    ############## IF USER SELECTS TO PULL FROM GRID ##################
    elif planet['type'] =='grid':
        try:
            db = create_engine('sqlite:///'+os.environ.get('FORTGRID_DIR'))
            header= pd.read_sql_table('header',db)
        except:
            raise Exception('Fortney Grid File Path is incorrect, or not initialized')

        #radius of star
        try:
            rstar = (star['radius']*u.Unit(star['r_unit'])).to(u.km)
        except:
            raise Exception("Radius of Star not supplied for scaling. Check exo_input['star']['radius']")

        #radius of planet
        try:
            rplan = (planet['radius']*u.Unit(planet['r_unit'])).to(u.km)
        except: 
            planet['radius'] = (1.25*c.R_jup).to(u.km)
            rplan = planet['radius']
            print('Default Planet Radius of 1.25 Rj given')

        #clouds 
        if planet['cloud'].find('flat') != -1: 
            planet['flat'] = int(planet['cloud'][4:])
            planet['ray'] = 0 
        elif planet['cloud'].find('ray') != -1: 
            planet['ray'] = int(planet['cloud'][3:])
            planet['flat'] = 0 
        elif int(planet['cloud']) == 0: 
            planet['flat'] = 0 
            planet['ray'] = 0     
        else:
            planet['flat'] = 0 
            planet['ray'] = 0 
            print('No cloud parameter not specified, default no clouds added')
        
        #chemistry 
        if planet['chem'] == 'noTiO': 
            planet['noTiO'] = True
            planet['eqchem'] = True 
        if planet['chem'] == 'eqchem': 
            planet['noTiO'] = False
            planet['eqchem'] = True 
            #grid does not allow clouds for cases with TiO
            planet['flat'] = 0 
            planet['ray'] = 0 

        #we are only using gravity of 25 and scaling by mass from there 
        fort_grav = 25.0*u.m/u.s/u.s
        df = header.loc[(header.gravity==fort_grav) & (header.temp==planet['temp'])
                           & (header.noTiO==planet['noTiO']) & (header.ray==planet['ray']) &
                           (header.flat==planet['flat'])]
        wave_planet=np.array(pd.read_sql_table(df['name'].values[0],db)['wavelength'])[::-1]

        r_lambda=np.array(pd.read_sql_table(df['name'].values[0],db)['radius'])*u.km
        z_lambda = r_lambda- (1.25*u.R_jup).to(u.km) #all fortney models have fixed 1.25 radii

        #scale with planetary mass 
        try:
            mass = (planet['mass']*u.Unit(planet['m_unit'])).to(u.kg)
            gravity = c.G*(mass)/(rplan.to(u.m))**2.0 #convert radius to m for gravity units
            #scale lambbda (this technically ignores the fact that scaleheight is altitude dependent)
            #therefore, it will not be valide for very very low gravities
            z_lambda = z_lambda*fort_grav/gravity
        except: 
            #keep original z lambda 
            gravity=25.0
            z_lambda = z_lambda*fort_grav/fort_grav
            print('Default Planet Gravity of 25 m/s2 given')  
        
        #create new wavelength dependent R based on scaled ravity
        r_lambda = z_lambda + rplan


        #finally compute (rp/r*)^2
        flux_planet = np.array(r_lambda**2/rstar**2)[::-1]
        planet['w_unit'] = 'um'
        planet['f_unit'] = 'rp^2/r*^2'
    else: 
        raise Exception("Incorrect Planet Type. Options are 'user','constant','grid'") 

    #Convert wave to micron 
    if planet['w_unit'] == 'um':
        wave_planet = wave_planet
    elif planet['w_unit'] == 'nm':
        wave_planet = wave_planet*1e-3
    elif planet['w_unit'] == 'cm':
        wave_planet = wave_planet*1e4
    elif planet['w_unit'] == 'Angs' :
        wave_planet = wave_planet*1e-4
    elif planet['w_unit'] == 'Hz' :
        wave_planet = 3e17/wave_planet
    elif planet['w_unit'] == 'sec' :
        wave_planet = wave_planet
    else: 
        raise Exception('Units are not correct. Pick um, nm, cm, Angs or sec.')

    if planet['w_unit'] == 'sec' :
        #star flux to feed into pandeia
        time = wave_planet
        flux_star = out_trans['flux_out_trans']
        wave_star = out_trans['wave']
        if planet['f_unit'] == 'fp/f*' :
            flux_planet = flux_planet 
        else: 
            print("Seconds with rp^2/r*^2 units not an option. Switch to Fp/F*")
            return 
        
        return {'time':time, 'wave':wave_star,'flux_out_trans':flux_star, 'planet_phase':flux_planet,
                'model_wave':time, 'model_spec':flux_planet, 'frac':(1.+flux_planet)}    
        
    else:
        #star flux to calc transit depth
        flux_star = out_trans['flux_out_trans']
        wave_star = out_trans['wave']

    
        #give them same wave min and wave max 
        wavemin = max([min(wave_planet), min(wave_star),0.5])
        wavemax = min([max(wave_planet),max(wave_star),15])
    
        flux_planet = flux_planet[(wave_planet>wavemin) & (wave_planet<wavemax)]
        wave_planet = wave_planet[(wave_planet>wavemin) & (wave_planet<wavemax)]
    
        flux_out_trans = np.interp(wave_planet, wave_star, flux_star)

        #convert to 1-depth 
        if planet['f_unit'] == 'rp^2/r*^2' :
            depth_fraction = 1.-flux_planet 
            flux_in_trans = depth_fraction*flux_out_trans
        elif planet['f_unit'] == 'fp/f*':
            depth_fraction = (1.0 + flux_planet)
            flux_in_trans = flux_out_trans*(1.0 + flux_planet)        
        else: 
            raise Exception('Units are not correct. Pick rp^2/r*^2 or fp/f*')
    
        results= {'wave':wave_planet, 'flux_in_trans': flux_in_trans, 'flux_out_trans':flux_out_trans,
                    'model_wave':wave_planet, 'model_spec': flux_planet, 'frac':depth_fraction} 
    return results

def hst_spec(planet,star) :
    """Calculates in transit flux 
    
    Takes output from `outTrans`, which is the normalized stellar flux, and 
    creates either a transit transmission spectrum, phase curve or emission spectrum. 
    Magnitude 
  
    Parameters 
    ----------
    planet: dict
        dictionary with direction to planet spectra, wavelength and flux units
    star: dict
        dictionary within exo_input with stellar information. Only 
        used when scaling Fortney Grid spectra to get (rp/r*)^2

    Return
    ------
    dict 
        dictionary with out of transit flux, in transit flux, original model 
        and corresponding wavelengths
    """ 
    
    if planet['type'] =='user':
        load_file = np.genfromtxt(planet['exopath'], dtype=(float, float), names='w, f')   
        #get wavelength 
        wave_planet = load_file['w']
        #get planet flux 
        flux_planet = load_file['f']
        
        #sort if not in ascending order 
        sort = np.array([wave_planet,flux_planet]).T
        sort= sort[sort[:,0].argsort()]
        wave_planet = sort[:,0]
        flux_planet = sort[:,1]
    
    ############## IF USER SELECTS CONSTANT VALUE ##################   
    elif planet['type'] == 'constant':
        rplan = (planet['radius']*u.Unit(planet['r_unit'])).to(u.km)
        rstar = (star['radius']*u.Unit(star['r_unit'])).to(u.km)

        #constant transit depth
        if planet['f_unit'] == 'rp^2/r*^2':
            wave_planet = np.linspace(0.1,3,500)
            planet['depth'] = float(rplan**2 / rstar**2)
            flux_planet = np.linspace(0.1,3,500)*0 + planet['depth']
            planet['w_unit'] = 'um'

        #constant fp/f* (using out_trans from user)
        elif planet['f_unit'] == 'fp/f*':
            planet['w_unit'] = 'um'
            wave_planet = np.linspace(0.1,3,500)
            flux_star = (bb.blackbody_nu(wave_planet*u.micron, star['temp']*u.K)*np.pi*u.sr).to(u.mJy)
            #MAKING SURE TO ADD IN SUPID PI FOR PER STERADIAN!!!!
            flux_planet = (bb.blackbody_nu(wave_planet*u.micron, planet['temp']*u.K)*np.pi*u.sr).to(u.mJy)
            # ( bb planet / pheonix sed ) * (rp/r*)^2
            flux_planet = np.array((flux_planet/flux_star) * (rplan/rstar)**2.0)


    ############## IF USER SELECTS TO PULL FROM GRID ##################
    elif planet['type'] =='grid':
        try:
            db = create_engine('sqlite:///'+os.environ.get('FORTGRID_DIR'))
            header= pd.read_sql_table('header',db)
        except:
            raise Exception('Fortney Grid File Path is incorrect, or not initialized')

        #radius of star
        try:
            rstar = (star['radius']*u.Unit(star['r_unit'])).to(u.km)
        except:
            raise Exception("Radius of Star not supplied for scaling. Check exo_input['star']['radius']")

        #radius of planet
        try:
            rplan = (planet['radius']*u.Unit(planet['r_unit'])).to(u.km)
        except: 
            planet['radius'] = (1.25*c.R_jup).to(u.km)
            rplan = planet['radius']
            print('Default Planet Radius of 1.25 Rj given')

        #clouds 
        if planet['cloud'].find('flat') != -1: 
            planet['flat'] = int(planet['cloud'][4:])
            planet['ray'] = 0 
        elif planet['cloud'].find('ray') != -1: 
            planet['ray'] = int(planet['cloud'][3:])
            planet['flat'] = 0 
        elif int(planet['cloud']) == 0: 
            planet['flat'] = 0 
            planet['ray'] = 0     
        else:
            planet['flat'] = 0 
            planet['ray'] = 0 
            print('No cloud parameter not specified, default no clouds added')
        
        #chemistry 
        if planet['chem'] == 'noTiO': 
            planet['noTiO'] = True
            planet['eqchem'] = True 
        if planet['chem'] == 'eqchem': 
            planet['noTiO'] = False
            planet['eqchem'] = True 
            #grid does not allow clouds for cases with TiO
            planet['flat'] = 0 
            planet['ray'] = 0 

        #we are only using gravity of 25 and scaling by mass from there 
        fort_grav = 25.0*u.m/u.s/u.s
        df = header.loc[(header.gravity==fort_grav) & (header.temp==planet['temp'])
                           & (header.noTiO==planet['noTiO']) & (header.ray==planet['ray']) &
                           (header.flat==planet['flat'])]
        wave_planet=np.array(pd.read_sql_table(df['name'].values[0],db)['wavelength'])[::-1]

        r_lambda=np.array(pd.read_sql_table(df['name'].values[0],db)['radius'])*u.km
        z_lambda = r_lambda- (1.25*u.R_jup).to(u.km) #all fortney models have fixed 1.25 radii

        #scale with planetary mass 
        try:
            mass = (planet['mass']*u.Unit(planet['m_unit'])).to(u.kg)
            gravity = c.G*(mass)/(rplan.to(u.m))**2.0 #convert radius to m for gravity units
            #scale lambbda (this technically ignores the fact that scaleheight is altitude dependent)
            #therefore, it will not be valide for very very low gravities
            z_lambda = z_lambda*fort_grav/gravity
        except: 
            #keep original z lambda 
            gravity=25.0
            z_lambda = z_lambda*fort_grav/fort_grav
            print('Default Planet Gravity of 25 m/s2 given')  
        
        #create new wavelength dependent R based on scaled ravity
        r_lambda = z_lambda + rplan


        #finally compute (rp/r*)^2
        flux_planet = np.array(r_lambda**2/rstar**2)[::-1]
        planet['w_unit'] = 'um'
        planet['f_unit'] = 'rp^2/r*^2'
    else: 
        raise Exception("Incorrect Planet Type. Options are 'user','constant','grid'") 

    #Convert wave to micron 
    if planet['w_unit'] == 'um':
        wave_planet = wave_planet
    elif planet['w_unit'] == 'nm':
        wave_planet = wave_planet*1e-3
    elif planet['w_unit'] == 'cm':
        wave_planet = wave_planet*1e4
    elif planet['w_unit'] == 'Angs' :
        wave_planet = wave_planet*1e-4
    elif planet['w_unit'] == 'Hz' :
        wave_planet = 3e17/wave_planet
    else: 
        raise Exception('Units are not correct. Pick um, nm, cm, Angs or sec.')

    
    return wave_planet, flux_planet



# ! /usr/bin/env pythonp
"""ramp effect model

Version 1.0.0: published version

Version 0.1: Adapted original IDL code (From D. Apai) to python by Yifan Zhou

"""
import numpy as np
import itertools


def RECTE(
        cRates,
        tExp,
        exptime=180,
        trap_pop_s=0,
        trap_pop_f=0,
        dTrap_s=50,
        dTrap_f=10,
        dt0=0,
        lost=0,
        mode='scanning'
):
    """This function calculates HST/WFC3/IR ramp effect profile based on
the charge trapping explanation developed in Zhou et al. (2017).

    Parameters
    ----------
    cRates : numpy.array
		intrinsic count rate of each exposures, unit: e/s

    tExp : numpy.array
		time stamps for the exposures, unit: seconds

    exptime : numpy.array or float
		(default 180 seconds) exposure time

    trap_pop_s : float or numpy.array
		(default 0) number of occupied slow population
        charge traps before the very beginning of the observation

	trap_pop_f: float or numpy.array
		(default 0) number of occupied fast population
        charge traps before the very beginning of the observation

    dTrap_s: float or numpy.array
		(default [0]) number of additional charges trapped
        by slow population traps during earth occultation

    dTrap_f : float or numpy.array
		(default [0]) number of additional charges trapped
        by fast population traps during earth occultation

    dt0: float
		(default 0) exposure time before the very beginning
        of the observation. It could be due to guidence adjustment

    lost: float
		(default 0, no lost) fraction of trapped electrons that are
        not eventually detected

    mode : string
		(default scanning, scanning or staring, or others),
        for scanning mode observation , the pixel no longer receive
        photons during the overhead time, in staring mode,
        the pixel keps receiving elctrons

	Returns
	-------
	numpy.array
		observed counts


    Example
	-------

    see Examples and Cookbook

    """
    nTrap_s = 1525.38  # 1320.0
    eta_trap_s = 0.013318  # 0.01311
    tau_trap_s = 1.63e4
    nTrap_f = 162.38
    eta_trap_f = 0.008407
    tau_trap_f = 281.463
    try:
        dTrap_f = itertools.cycle(dTrap_f)
        dTrap_s = itertools.cycle(dTrap_s)
        dt0 = itertools.cycle(dt0)
    except TypeError:
        dTrap_f = itertools.cycle([dTrap_f])
        dTrap_s = itertools.cycle([dTrap_s])
        dt0 = itertools.cycle([dt0])
    obsCounts = np.zeros(len(tExp))
    trap_pop_s = min(trap_pop_s, nTrap_s)
    trap_pop_f = min(trap_pop_f, nTrap_f)
    for i in range(len(tExp)):
        try:
            dt = tExp[i+1] - tExp[i]
        except IndexError:
            dt = exptime
        f_i = cRates[i]
        c1_s = eta_trap_s * f_i / nTrap_s + 1 / tau_trap_s  # a key factor
        c1_f = eta_trap_f * f_i / nTrap_f + 1 / tau_trap_f
        # number of trapped electron during one exposure
        dE1_s = (eta_trap_s * f_i / c1_s - trap_pop_s) * \
            (1 - np.exp(-c1_s * exptime))
        dE1_f = (eta_trap_f * f_i / c1_f - trap_pop_f) * \
            (1 - np.exp(-c1_f * exptime))
        dE1_s = min(trap_pop_s + dE1_s, nTrap_s) - trap_pop_s
        dE1_f = min(trap_pop_f + dE1_f, nTrap_f) - trap_pop_f
        trap_pop_s = min(trap_pop_s + dE1_s, nTrap_s)
        trap_pop_f = min(trap_pop_f + dE1_f, nTrap_f)
        obsCounts[i] = f_i * exptime - dE1_s - dE1_f
        if dt < 5 * exptime:  # whether next exposure is in next batch of exposures
            # same orbits
            if mode == 'scanning':
                # scanning mode, no incoming flux between exposures
                dE2_s = - trap_pop_s * (1 - np.exp(-(dt - exptime)/tau_trap_s))
                dE2_f = - trap_pop_f * (1 - np.exp(-(dt - exptime)/tau_trap_f))
            elif mode == 'staring':
                # for staring mode, there is flux between exposures
                dE2_s = (eta_trap_s * f_i / c1_s - trap_pop_s) * \
                    (1 - np.exp(-c1_s * (dt - exptime)))
                dE2_f = (eta_trap_f * f_i / c1_f - trap_pop_f) * \
                    (1 - np.exp(-c1_f * (dt - exptime)))
            else:
                # others, same as scanning
                dE2_s = - trap_pop_s * (1 - np.exp(-(dt - exptime)/tau_trap_s))
                dE2_f = - trap_pop_f * (1 - np.exp(-(dt - exptime)/tau_trap_f))
            trap_pop_s = min(trap_pop_s + dE2_s, nTrap_s)
            trap_pop_f = min(trap_pop_f + dE2_f, nTrap_f)
        elif dt < 1200:
            # considering in orbit download scenario
            trap_pop_s = min(
                trap_pop_s * np.exp(-(dt-exptime)/tau_trap_s), nTrap_s)
            trap_pop_f = min(
                trap_pop_f * np.exp(-(dt-exptime)/tau_trap_f), nTrap_f)
        else:
            # switch orbit
            dt0_i = next(dt0)
            trap_pop_s = min(trap_pop_s * np.exp(-(dt-exptime-dt0_i)/tau_trap_s) +
                             next(dTrap_s), nTrap_s)
            trap_pop_f = min(trap_pop_f * np.exp(-(dt-exptime-dt0_i)/tau_trap_f) +
                             next(dTrap_f), nTrap_f)
            f_i = cRates[i + 1]
            c1_s = eta_trap_s * f_i / nTrap_s + 1 / tau_trap_s  # a key factor
            c1_f = eta_trap_f * f_i / nTrap_f + 1 / tau_trap_f
            dE3_s = (eta_trap_s * f_i / c1_s - trap_pop_s) * \
                (1 - np.exp(-c1_s * dt0_i))
            dE3_f = (eta_trap_f * f_i / c1_f - trap_pop_f) * \
                (1 - np.exp(-c1_f * dt0_i))
            dE3_s = min(trap_pop_s + dE3_s, nTrap_s) - trap_pop_s
            dE3_f = min(trap_pop_f + dE3_f, nTrap_f) - trap_pop_f
            trap_pop_s = min(trap_pop_s + dE3_s, nTrap_s)
            trap_pop_f = min(trap_pop_f + dE3_f, nTrap_f)
        trap_pop_s = max(trap_pop_s, 0)
        trap_pop_f = max(trap_pop_f, 0)

    return obsCounts


if __name__ == '__main__':
    pass



from bokeh.plotting import show, Figure
from bokeh.io import output_file as outputfile
from bokeh.io import output_notebook  as outnotebook
import pickle as pk
import numpy as np
from bokeh.layouts import row
import pandas as pd
def jwst_1d_spec(result_dict, model=True, title='Model + Data + Error Bars', output_file = 'data.html',legend = False,
        R=False,  num_tran = False, plot_width=800, plot_height=400,x_range=[1,10],y_range=None, plot=True,
        output_notebook=False):
    """Plots 1d simulated spectrum and rebin or rescale for more transits

    Plots 1d data points with model in the background (if wanted). Designed to read in exact
    output of run_pandexo.

    Parameters
    ----------
    result_dict : dict or list of dict
        Dictionary from pandexo output. If parameter space was run in run_pandexo
        make sure to restructure the input as a list of dictionaries without they key words
        that run_pandexo assigns.
    model : bool
        (Optional) True is default. True plots model, False does not plot model
    title : str
        (Optional) Title of plot. Default is "Model + Data + Error Bars".
    output_file : str
        (Optional) name of html file for you bokeh plot. After bokeh plot is rendered you will
        have the option to save as png.
    legend : bool
        (Optional) Default is False. True, plots legend.
    R : float
        (Optional) Rebin data from native instrument resolution to specified resolution. Dafult is False,
        no binning. Here I adopt R as w[1]/(w[2] - w[0]) to maintain consistency with `pandeia.engine`
    num_tran : float
        (Optional) Scales data by number of transits to improve error by sqrt(`num_trans`)
    plot_width : int
        (Optional) Sets the width of the plot. Default = 800
    plot_height : int
        (Optional) Sets the height of the plot. Default = 400
    y_range : list of int
        (Optional) sets y range of plot. Defaut is +- 10% of max and min
    x_range : list of int
        (Optional) Sets x range of plot. Default = [1,10]
    plot : bool
        (Optional) Supresses the plot if not wanted (Default = True)
    out_notebook : bool 
        (Optional) Output notebook. Default is false, if true, outputs in the notebook

    Returns
    -------
    x,y,e : list of arrays
        Returns wave axis, spectrum and associated error in list format. x[0] will be correspond
        to the first dictionary input, x[1] to the second, etc.

    Examples
    --------

    >>> jwst_1d_spec(result_dict, num_tran = 3, R = 35) #for a single plot

    If you wanted to save each of the axis that were being plotted:

    >>> x,y,e = jwst_1d_data([result_dict1, result_dict2], model=False, num_tran = 5, R = 100) #for multiple

    See Also
    --------
    jwst_noise, jwst_1d_bkg, jwst_1d_flux, jwst_1d_snr, jwst_2d_det, jwst_2d_sat

    """
    outx=[]
    outy=[]
    oute=[]
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"
    if output_notebook & plot:
        outnotebook()
    elif plot:
        outputfile(output_file)
    colors = ['black','blue','red','orange','yellow','purple','pink','cyan','grey','brown']
    #make sure its iterable
    if type(result_dict) != list:
        result_dict = [result_dict]

    if type(legend)!=bool:
        legend_keys = legend
        legend = True
        if type(legend_keys) != list:
            legend_keys = [legend_keys]

    i = 0
    for dictt in result_dict:
        ntran_old = dictt['timing']['Number of Transits']
        to = dictt['timing']["Num Integrations Out of Transit"]
        ti = dictt['timing']["Num Integrations In Transit"]
        #remove any nans
        y = dictt['FinalSpectrum']['spectrum_w_rand']
        x = dictt['FinalSpectrum']['wave'][~np.isnan(y)]
        err = dictt['FinalSpectrum']['error_w_floor'][~np.isnan(y)]
        y = y[~np.isnan(y)]


        if (R == False) & (num_tran == False):
            x=x
            y=y
        elif (R != False) & (num_tran != False):
            new_wave = bin_wave_to_R(x, R)
            out = uniform_tophat_sum(new_wave,x, dictt['RawData']['electrons_out']*num_tran/ntran_old)
            inn = uniform_tophat_sum(new_wave,x, dictt['RawData']['electrons_in']*num_tran/ntran_old)
            vout = uniform_tophat_sum(new_wave,x, dictt['RawData']['var_out']*num_tran/ntran_old)
            vin = uniform_tophat_sum(new_wave,x, dictt['RawData']['var_in']*num_tran/ntran_old)
            var_tot = (to/ti/out)**2.0 * vin + (inn*to/ti/out**2.0)**2.0 * vout
            if dictt['input']['Primary/Secondary']=='fp/f*':
                fac = -1.0
            else:
                fac = 1.0
            rand_noise = np.sqrt((var_tot))*(np.random.randn(len(new_wave)))
            raw_spec = (out/to-inn/ti)/(out/to)
            sim_spec = fac*(raw_spec + rand_noise )
            x = new_wave
            y = sim_spec
            err = np.sqrt(var_tot)
        elif (R == False) & (num_tran != False):
            out = dictt['RawData']['electrons_out']*num_tran/ntran_old
            inn = dictt['RawData']['electrons_in']*num_tran/ntran_old
            vout = dictt['RawData']['var_out']*num_tran/ntran_old
            vin = dictt['RawData']['var_in']*num_tran/ntran_old
            var_tot = (to/ti/out)**2.0 * vin + (inn*to/ti/out**2.0)**2.0 * vout
            if dictt['input']['Primary/Secondary']=='fp/f*':
                fac = -1.0
            else:
                fac = 1.0
            rand_noise = np.sqrt((var_tot))*(np.random.randn(len(x)))
            raw_spec = (out/to-inn/ti)/(out/to)
            sim_spec = fac*(raw_spec + rand_noise )
            x = x
            y = sim_spec
            err = np.sqrt(var_tot)
        elif (R != False) & (num_tran == False):
            new_wave = bin_wave_to_R(x, R)
            out = uniform_tophat_sum(new_wave,x, dictt['RawData']['electrons_out'])
            inn = uniform_tophat_sum(new_wave,x, dictt['RawData']['electrons_in'])
            vout = uniform_tophat_sum(new_wave,x, dictt['RawData']['var_out'])
            vin = uniform_tophat_sum(new_wave,x, dictt['RawData']['var_in'])
            var_tot = (to/ti/out)**2.0 * vin + (inn*to/ti/out**2.0)**2.0 * vout
            if dictt['input']['Primary/Secondary']=='fp/f*':
                fac = -1.0
            else:
                fac = 1.0
            rand_noise = np.sqrt((var_tot))*(np.random.randn(len(new_wave)))
            raw_spec = (out/to-inn/ti)/(out/to)
            sim_spec = fac*(raw_spec + rand_noise )
            x = new_wave
            y = sim_spec
            err = np.sqrt(var_tot)
        else:
            print("Something went wrong. Cannot enter both resolution and ask to bin to new wave")
            return

        #create error bars for Bokeh's multi_line and drop nans
        data = pd.DataFrame({'x':x, 'y':y,'err':err}).dropna()

        y_err = []
        x_err = []
        for px, py, yerr in zip(data['x'], data['y'], data['err']):
            np.array(x_err.append((px, px)))
            np.array(y_err.append((py - yerr, py + yerr)))
        #initialize Figure
        if i == 0:
            #Define units for x and y axis
            y_axis_label = dictt['input']['Primary/Secondary']

            if y_axis_label == 'fp/f*': p = -1.0
            else: y_axis_label = y_axis_label

            if dictt['input']['Calculation Type'] =='phase_spec':
                x_axis_label='Time (secs)'
                x_range = [min(x), max(x)]
            else:
                x_axis_label='Wavelength [microns]'

            if y_range!=None:
                ylims = y_range
            else:
                ylims = [min(dictt['OriginalInput']['model_spec'])- 0.1*min(dictt['OriginalInput']['model_spec']),
                 0.1*max(dictt['OriginalInput']['model_spec'])+max(dictt['OriginalInput']['model_spec'])]

            fig1d = Figure(x_range=x_range, y_range = ylims,
               plot_width = plot_width, plot_height =plot_height,title=title,x_axis_label=x_axis_label,
              y_axis_label = y_axis_label, tools=TOOLS, background_fill_color = 'white')


        #plot model, data, and errors
        if model:
            mxx = dictt['OriginalInput']['model_wave']
            myy = dictt['OriginalInput']['model_spec']
            my = uniform_tophat_mean(x, mxx,myy)
            model_line = pd.DataFrame({'x':x, 'my':my}).dropna()
            fig1d.line(model_line['x'],model_line['my'], color=colors[i],alpha=0.2, line_width = 4)


        if legend:
            fig1d.circle(data['x'], data['y'], color=colors[i], legend = legend_keys[i])
        else:
            fig1d.circle(data['x'], data['y'], color=colors[i])
        outx += [data['x'].values]
        outy += [data['y'].values]
        oute += [data['err'].values]
        fig1d.multi_line(x_err, y_err,color=colors[i])
        i += 1
    if plot:
        show(fig1d)
    return outx,outy,oute


def bin_wave_to_R(w, R):
    """Creates new wavelength axis at specified resolution

    Parameters
    ----------
    w : list of float or numpy array of float
        Wavelength axis to be rebinned
    R : float or int
        Resolution to bin axis to

    Returns
    -------
    list of float
        New wavelength axis at specified resolution

    Examples
    --------

    >>> newwave = bin_wave_to_R(np.linspace(1,2,1000), 10)
    >>> print(len(newwave))
    11
    """
    wave = []
    tracker = min(w)
    i = 1
    ind= 0
    firsttime = True
    while(tracker<max(w)):
        if i <len(w)-1:
            dlambda = w[i]-w[ind]
            newR = w[i]/dlambda
            if (newR < R) & (firsttime):
                tracker = w[ind]
                wave += [tracker]
                ind += 1
                i += 1
                firsttime = True
            elif newR < R:
                tracker = w[ind]+dlambda/2.0
                wave +=[tracker]
                ind = (np.abs(w-tracker)).argmin()
                i = ind+1
                firsttime = True
            else:
                firsttime = False
                i+=1
        else:
            tracker = max(w)
            wave += [tracker]
    return wave

def uniform_tophat_sum(xnew,x, y):
    """Adapted from Mike R. Line to rebin spectra

    Takes sum of group of points in bin of wave points
    Parameters
    ----------
    xnew : list of float or numpy array of float
        New wavelength grid to rebin to
    x : list of float or numpy array of float
        Old wavelength grid to get rid of
    y : list of float or numpy array of float
        New rebinned y axis

    Returns
    -------
    array of floats
        new y axis

    Examples
    --------

    >>> oldgrid = np.linspace(1,3,100)
    >>> y = np.zeros(100)+10.0
    >>> newy = uniform_tophat_sum(np.linspace(2,3,3), oldgrid, y)
    >>> newy
    array([ 240.,  250.,  130.])
    """
    xnew = np.array(xnew)
    szmod=xnew.shape[0]
    delta=np.zeros(szmod)
    ynew=np.zeros(szmod)
    delta[0:-1]=xnew[1:]-xnew[:-1]
    delta[szmod-1]=delta[szmod-2]
    #pdb.set_trace()
    for i in range(szmod-1):
        i=i+1
        loc=np.where((x >= xnew[i]-0.5*delta[i-1]) & (x < xnew[i]+0.5*delta[i]))
        ynew[i]=np.sum(y[loc])
    loc=np.where((x > xnew[0]-0.5*delta[0]) & (x < xnew[0]+0.5*delta[0]))
    ynew[0]=np.sum(y[loc])
    return ynew

def uniform_tophat_mean(xnew,x, y):
    """Adapted from Mike R. Line to rebin spectra

    Takes average of group of points in bin

    Parameters
    ----------
    xnew : list of float or numpy array of float
        New wavelength grid to rebin to
    x : list of float or numpy array of float
        Old wavelength grid to get rid of
    y : list of float or numpy array of float
        New rebinned y axis

    Returns
    -------
    array of floats
        new y axis

    Examples
    --------
    >>> oldgrid = np.linspace(1,3,100)
    >>> y = np.zeros(100)+10.0
    >>> newy = uniform_tophat_sum(np.linspace(2,3,3), oldgrid, y)
    >>> newy
    array([ 240.,  250.,  130.])
    """
    xnew = np.array(xnew)
    szmod=xnew.shape[0]
    delta=np.zeros(szmod)
    ynew=np.zeros(szmod)
    delta[0:-1]=xnew[1:]-xnew[:-1]
    delta[szmod-1]=delta[szmod-2]
    #pdb.set_trace()
    for i in range(szmod-1):
        i=i+1
        loc=np.where((x >= xnew[i]-0.5*delta[i-1]) & (x < xnew[i]+0.5*delta[i]))
        ynew[i]=np.mean(y[loc])
    loc=np.where((x > xnew[0]-0.5*delta[0]) & (x < xnew[0]+0.5*delta[0]))
    ynew[0]=np.mean(y[loc])
    return ynew

def jwst_1d_flux(result_dict, plot=True, output_file= 'flux.html'):
    """Plot flux rate in e/s

    Parameters
    ----------
    result_dict : dict
        Dictionary from pandexo output. If parameter space was run in run_pandexo
        make sure to restructure the input as a list of dictionaries without they key words
        that run_pandexo assigns.
    plot : bool
        (Optional) True renders plot, Flase does not. Default=True
    output_file : str
        (Optional) Default = 'flux.html'
    Return
    ------
    x : numpy array
        micron
    y : numpy array
        1D flux rate in electrons/s

    See Also
    --------
    jwst_1d_spec, jwst_1d_bkg, jwst_noise, jwst_1d_snr, jwst_2d_det, jwst_2d_sat
    """
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"
    out = result_dict['PandeiaOutTrans']

    # Flux 1d
    x, y = out['1d']['extracted_flux']
    x = x[~np.isnan(y)]
    y = y[~np.isnan(y)]

    plot_flux_1d1 = Figure(tools=TOOLS,
                         x_axis_label='Wavelength [microns]',
                         y_axis_label='Flux (e/s)', title="Out of Transit Flux Rate",
                         plot_width=800, plot_height=300)
    plot_flux_1d1.line(x, y, line_width = 4, alpha = .7)

    if plot:
        outputfile(output_file)
        show(plot_flux_1d1)
    return x,y

def jwst_1d_snr(result_dict, plot=True, output_file='snr.html'):
    """Plot SNR

    Parameters
    ----------
    result_dict : dict
        Dictionary from pandexo output. If parameter space was run in run_pandexo
        make sure to restructure the input as a list of dictionaries without they key words
        that run_pandexo assigns.
    plot : bool
        (Optional) True renders plot, Flase does not. Default=True
    output_file : str
        (Optional) Default = 'snr.html'

    Return
    ------
    x : numpy array
        micron
    y : numpy array
        1D SNR

    See Also
    --------
    jwst_1d_bkg, jwst_noise, jwst_1d_flux, jwst_1d_spec, jwst_2d_det, jwst_2d_sat
    """
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"
    # Flux 1d
    x= result_dict['RawData']['wave']
    electrons_out = result_dict['RawData']['electrons_out']
    y = electrons_out/np.sqrt(result_dict['RawData']['var_out'])
    x = x[~np.isnan(y)]
    y = y[~np.isnan(y)]
    plot_snr_1d1 = Figure(tools=TOOLS,
                         x_axis_label='Wavelength (micron)',
                         y_axis_label='SNR', title="SNR Out of Trans",
                         plot_width=800, plot_height=300)
    plot_snr_1d1.line(x, y, line_width = 4, alpha = .7)
    if plot:
        outputfile(output_file)
        show(plot_snr_1d1)
    return x,y

def jwst_1d_bkg(result_dict, plot=True, output_file='bkg.html'):
    """Plot background

    Parameters
    ----------
    result_dict : dict
        Dictionary from pandexo output. If parameter space was run in run_pandexo
        make sure to restructure the input as a list of dictionaries without they key words
        that run_pandexo assigns.
    plot : bool
        (Optional) True renders plot, Flase does not. Default=True
    output_file : str
        (Optional) Default = bkt.html

    Return
    ------
    x : numpy array
        micron
    y : numpy array
        1D bakground e/s

    See Also
    --------
    jwst_1d_spec, jwst_noise, jwst_1d_flux, jwst_1d_snr, jwst_2d_det, jwst_2d_sat
    """
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"
    # BG 1d
    out = result_dict['PandeiaOutTrans']
    x, y = out['1d']['extracted_bg_only']
    y = y[~np.isnan(y)]
    x = x[~np.isnan(y)]
    plot_bg_1d1 = Figure(tools=TOOLS,
                         x_axis_label='Wavelength [microns]',
                         y_axis_label='Flux (e/s)', title="Background",
                         plot_width=800, plot_height=300)
    plot_bg_1d1.line(x, y, line_width = 4, alpha = .7)
    if plot:
        outputfile(output_file)
        show(plot_bg_1d1)
    return x,y

def jwst_noise(result_dict, plot=True, output_file= 'noise.html'):
    """Plot background

    Parameters
    ----------
    result_dict : dict
        Dictionary from pandexo output. If parameter space was run in run_pandexo
        make sure to restructure the input as a list of dictionaries without they key words
        that run_pandexo assigns.
    plot : bool
        (Optional) True renders plot, Flase does not. Default=True
    output_file : str
        (Optional) Default = 'noise.html'
    Return
    ------
    x : numpy array
        micron
    y : numpy array
        1D noise (ppm)

    See Also
    --------
    jwst_1d_spec, jwst_1d_bkg, jwst_1d_flux, jwst_1d_snr, jwst_2d_det, jwst_2d_sat
    """
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"    #saturation

    x = result_dict['FinalSpectrum']['wave']
    y = result_dict['FinalSpectrum']['error_w_floor']*1e6
    x = x[~np.isnan(y)]
    y = y[~np.isnan(y)]
    ymed = np.median(y)


    plot_noise_1d1 = Figure(tools=TOOLS,#responsive=True,
                         x_axis_label='Wavelength (micron)',
                         y_axis_label='Error on Spectrum (PPM)', title="Error Curve",
                         plot_width=800, plot_height=300, y_range = [0,2.0*ymed])
    ymed = np.median(y)
    plot_noise_1d1.circle(x, y, line_width = 4, alpha = .7)
    if plot:
        outputfile(output_file)
        show(plot_noise_1d1)
    return x,y

def jwst_2d_det(result_dict, plot=True, output_file='det2d.html'):
    """Plot 2d detector image

    Parameters
    ----------
    result_dict : dict
        Dictionary from pandexo output. If parameter space was run in run_pandexo
        make sure to restructure the input as a list of dictionaries without they key words
        that run_pandexo assigns.

    plot : bool
        (Optional) True renders plot, Flase does not. Default=True
    output_file : str
        (Optional) Default = 'det2d.html'

    Return
    ------
    numpy array
        2D array of out of transit detector simulation
    See Also
    --------
    jwst_1d_spec, jwst_1d_bkg, jwst_1d_flux, jwst_1d_snr, jwst_noise, jwst_2d_sat

    """
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"
    out = result_dict['PandeiaOutTrans']
    data = out['2d']['detector']


    xr, yr = data.shape

    plot_detector_2d = Figure(tools="pan,wheel_zoom,box_zoom,reset,hover,save",
                         x_range=[0, yr], y_range=[0, xr],
                         x_axis_label='Pixel', y_axis_label='Spatial',
                         title="2D Detector Image",
                        plot_width=800, plot_height=300)

    plot_detector_2d.image(image=[data], x=[0], y=[0], dh=[xr], dw=[yr],
                      palette="Spectral11")
    if plot:
        outputfile(output_file)
        show(plot_detector_2d)
    return data


def jwst_2d_sat(result_dict, plot=True, output_file='sat2d.html'):
    """Plot 2d saturation profile

    Parameters
    ----------
    result_dict : dict
        Dictionary from pandexo output. If parameter space was run in run_pandexo
        make sure to restructure the input as a list of dictionaries without they key words
        that run_pandexo assigns.

    plot : bool
        (Optional) True renders plot, Flase does not. Default=True
    output_file : str
        (Optional) Default = 'sat2d.html'

    Return
    ------
    numpy array
        2D array of out of transit detector simulation

    See Also
    --------
    jwst_1d_spec, jwst_1d_bkg, jwst_1d_flux, jwst_1d_snr, jwst_2d_det, jwst_noise
    """
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"    #saturation
    out = result_dict['PandeiaOutTrans']
    data = out['2d']['saturation']
    xr, yr = data.shape
    plot_sat_2d = Figure(tools=TOOLS,
                         x_range=[0, yr], y_range=[0, xr],
                         x_axis_label='Pixel', y_axis_label='Spatial',
                         title="Saturation",
                        plot_width=800, plot_height=300)

    plot_sat_2d.image(image=[data], x=[0], y=[0], dh=[xr], dw=[yr],
                      palette="Spectral11")
    if plot:
        outputfile(output_file)
        show(plot_sat_2d)
    return data

def hst_spec(result_dict, plot=True, output_file ='hstspec.html', model = True):
    """Plot 1d spec with error bars for hst

    Parameters
    ----------
    result_dict : dict
        Dictionary from pandexo output.

    plot : bool
        (Optional) True renders plot, False does not. Default=True
    model : bool
        (Optional) Plot model under data. Default=True
    output_file : str
        (Optional) Default = 'hstspec.html'

    Return
    ------
    x : numpy array
        micron
    y : numpy array
        1D spec fp/f* or rp^2/r*^2
    e : numpy array
        1D rms noise
    modelx : numpy array
        micron
    modely : numpy array
        1D spec fp/f* or rp^2/r*^2
    See Also
    --------
    hst_time
    """
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"
    #plot planet spectrum
    mwave = result_dict['planet_spec']['model_wave']
    mspec = result_dict['planet_spec']['model_spec']

    binwave = result_dict['planet_spec']['binwave']
    binspec = result_dict['planet_spec']['binspec']

    error = result_dict['planet_spec']['error']
    error = np.zeros(len(binspec))+ error
    xlims = [result_dict['planet_spec']['wmin'], result_dict['planet_spec']['wmax']]
    ylims = [np.min(binspec)-2.0*error[0], np.max(binspec)+2.0*error[0]]

    plot_spectrum = Figure(plot_width=800, plot_height=300, x_range=xlims,
                               y_range=ylims, tools=TOOLS,#responsive=True,
                                 x_axis_label='Wavelength [microns]',
                                 y_axis_label='Ratio',
                               title="Original Model with Observation")

    y_err = []
    x_err = []
    for px, py, yerr in zip(binwave, binspec, error):
        np.array(x_err.append((px, px)))
        np.array(y_err.append((py - yerr, py + yerr)))
    if model:
        plot_spectrum.line(mwave,mspec, color= "black", alpha = 0.5, line_width = 4)
    plot_spectrum.circle(binwave,binspec, line_width=3, line_alpha=0.6)
    plot_spectrum.multi_line(x_err, y_err)

    if plot:
        outputfile(output_file)
        show(plot_spectrum)

    return binwave, binspec, error, mwave, mspec

def hst_time(result_dict, plot=True, output_file ='hsttime.html', model = True):
    """Plot earliest and latest start times for hst observation

    Parameters
    ----------
    result_dict : dict
        Dictionary from pandexo output.

    plot : bool
        (Optional) True renders plot, False does not. Default=True
    model : bool
        (Optional) Plot model under data. Default=True
    output_file : str
        (Optional) Default = 'hsttime.html'

    Return
    ------
    obsphase1 : numpy array
        earliest start time
    obstr1 : numpy array
        white light curve
    obsphase2 : numpy array
        latest start time
    obstr2 : numpy array
        white light curve
    rms : numpy array
        1D rms noise

    See Also
    --------
    hst_spec
    """
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"
    #earliest and latest start times
    obsphase1 = result_dict['calc_start_window']['obsphase1']
    obstr1 = result_dict['calc_start_window']['obstr1']
    rms = result_dict['calc_start_window']['light_curve_rms']
    obsphase2 = result_dict['calc_start_window']['obsphase2']
    obstr2 = result_dict['calc_start_window']['obstr2']
    phase1 = result_dict['calc_start_window']['phase1']
    phase2 = result_dict['calc_start_window']['phase2']
    trmodel1 = result_dict['calc_start_window']['trmodel1']
    trmodel2 = result_dict['calc_start_window']['trmodel2']

    if isinstance(rms, float):
        rms = np.zeros(len(obsphase1))+rms
    y_err1 = []
    x_err1 = []
    for px, py, yerr in zip(obsphase1, obstr1, rms):
        np.array(x_err1.append((px, px)))
        np.array(y_err1.append((py - yerr, py + yerr)))

    y_err2 = []
    x_err2 = []
    for px, py, yerr in zip(obsphase2, obstr2, rms):
        np.array(x_err2.append((px, px)))
        np.array(y_err2.append((py - yerr, py + yerr)))

    early = Figure(plot_width=400, plot_height=300,
                               tools=TOOLS,#responsive=True,
                                 x_axis_label='Orbital Phase',
                                 y_axis_label='Flux',
                               title="Earliest Start Time")

    if model: early.line(phase1, trmodel1, color='black',alpha=0.5, line_width = 4)
    early.circle(obsphase1, obstr1, line_width=3, line_alpha=0.6)
    early.multi_line(x_err1, y_err1)

    late = Figure(plot_width=400, plot_height=300,
                                tools=TOOLS,#responsive=True,
                                 x_axis_label='Orbital Phase',
                                 y_axis_label='Flux',
                               title="Latest Start Time")
    if model: late.line(phase2, trmodel2, color='black',alpha=0.5, line_width = 3)
    late.circle(obsphase2, obstr2, line_width=3, line_alpha=0.6)
    late.multi_line(x_err2, y_err2)

    start_time = row(early, late)

    if plot:
        outputfile(output_file)
        show(start_time)



    return obsphase1, obstr1, obsphase2, obstr2,rms


def hst_simulated_lightcurve(result_dict, plot=True, output_file ='hsttime.html', model = True):
    """Plot simulated HST light curves (in fluece) for earliest and latest start times

    Parameters
    ----------
    result_dict : dict
        Dictionary from pandexo output.

    plot : bool
        (Optional) True renders plot, False does not. Default=True
    model : bool
        (Optional) Plot model under data. Default=True
    output_file : str
        (Optional) Default = 'hsttime.html'

    Return
    ------
    obsphase1 : numpy array
        earliest start time
    counts1 : numpy array
        white light curve in fluence (e/pixel)
    obsphase2 : numpy array
        latest start time
    counts2 : numpy array
        white light curve in fluence (e/pixel)
    rms : numpy array
        1D rms noise

    See Also
    --------
    hst_spec
    """
    TOOLS = "pan,wheel_zoom,box_zoom,reset,save"
    # earliest and latest start times
    obsphase1 = result_dict['light_curve']['obsphase1']
    rms = result_dict['light_curve']['light_curve_rms']
    obsphase2 = result_dict['light_curve']['obsphase2']
    phase1 = result_dict['light_curve']['phase1']
    phase2 = result_dict['light_curve']['phase2']
    counts1 = result_dict['light_curve']['counts1']
    counts2 = result_dict['light_curve']['counts2']
    count_noise = result_dict['light_curve']['count_noise']
    ramp_included = result_dict['light_curve']['ramp_included']
    model_counts1 = result_dict['light_curve']['model_counts1']
    model_counts2 = result_dict['light_curve']['model_counts2']

    if isinstance(count_noise, float):
        rms = np.zeros(len(counts1)) + count_noise
    y_err1 = []
    x_err1 = []
    for px, py, yerr in zip(obsphase1, counts1, rms):
        np.array(x_err1.append((px, px)))
        np.array(y_err1.append((py - yerr, py + yerr)))

    y_err2 = []
    x_err2 = []
    for px, py, yerr in zip(obsphase2, counts2, rms):
        np.array(x_err2.append((px, px)))
        np.array(y_err2.append((py - yerr, py + yerr)))

    if ramp_included:
        title_description = " (Ramp Included)"
    else:
        title_description =" (Ramp Removed)"

    early = Figure(plot_width=400, plot_height=300,
                               tools=TOOLS,#responsive=True,
                                 x_axis_label='Orbital Phase',
                                 y_axis_label='Flux [electrons/pixel]',
                               title="Earliest Start Time" + title_description)

    if model:
        early.line(phase1, model_counts1, color='black', alpha=0.5, line_width=4)
    early.circle(obsphase1, counts1, line_width=3, line_alpha=0.6)
    early.multi_line(x_err1, y_err1)

    late = Figure(plot_width=400, plot_height=300,
                  tools=TOOLS,  # responsive=True,
                  x_axis_label='Orbital Phase',
                  y_axis_label='Flux [electrons/pixel]',
                  title="Latest Start Time" + title_description)
    if model:
        late.line(phase2, model_counts2, color='black', alpha=0.5, line_width=3)
    late.circle(obsphase2, counts2, line_width=3, line_alpha=0.6)
    late.multi_line(x_err2, y_err2)

    start_time = row(early, late)

    if plot:
        outputfile(output_file)
        show(start_time)

    return obsphase1, counts1, obsphase2, counts2, rms



import numpy as np

class ExtractSpec():
    """Different methods for computing noise. 
    
    PandExo has several different methods of computing noise.  
    MULTIACCUM (slope method) assumses that each frame is fit for up the ramp and 
    that the final nosie includes correlated noise from fitting each frame up the ramp. 
    First Minus Last assumes that intermediate frames are not used and final noise is just 
    the first frame minus the last frame. 2d extract uses Pandeia's 2d simualtions to 
    extracts the spectrum from the 2d extraction box. It extracts the entire postage stamp 
    so it might be an overestimate of flux (in contrast pandeia requires background 
    extraction region to be at least equal to the flux extraction region)
    
    Noise Components Included: 
        - Shot
        - Background and Dark Current
        - Read noise 
        
    Parameters
    ----------
    inn : dict
        In transit dictionary computed from PandExo 
    out : dict
        Out of transit dictionary comptued from PandExo 
    rn : float
        Read noise electrons 
    extraction_area : float
        Number of extracted pixels (square pixels) 
    timing : dict  
        Dictionary of computed JWST timing products 
    
    Methods
    -------
    loopingL
        extracts pixels from center to bottom
    loopingU
        extracts pixels from center to top
    sum_spatial
        sums pixels in optimal extraction region
    extract_region
        determines optimal extraction region
    run_2d_extract
        top level to extract spec from 2d pandeia output
    run_slope_method
        computes noise using multiaccum formulation
    run_f_minus_l
        computs noise using first minus last
    run_phase_spec
        computs noise for phase curve observations     
    """
    def __init__(self, inn, out, rn, extraction_area, timing):
        self.inn = inn
        self.out = out 
        self.ngroups_per_int = timing["APT: Num Groups per Integration"]
        self.nint_out = timing["Num Integrations Out of Transit"]
        self.nint_in = timing["Num Integrations In Transit"]
        self.tframe = timing["Seconds per Frame"]
        self.rn = rn 
        self.extraction_area = extraction_area

        #on source out versus in 
        self.exptime_per_int = self.tframe * (self.ngroups_per_int-1.0)

        self.on_source_in = self.tframe * (self.ngroups_per_int-1.0) * self.nint_in
        self.on_source_out = self.tframe * (self.ngroups_per_int-1.0) * self.nint_out

    def loopingL(self, cen, signal_col, noise_col, bkg_col):
        """Finds bottom of the optimal extraction region.
    
        Find location where SNR is the highest and loop from highest value downward 
    
        Parameters
        ----------
        cen : float or int 
            Pixel where SNR is the highest 
        signal_col : array of float 
            Array of fluxes to be extracted in a single column on the detector 
        noise_col : array of float 
            Array of noise to be extracted in a single column on the detector 
        bkg_col : array of float 
            Array of background fluxes to be extracted in a single column on the detector 
    
        Return 
        ------
        int
            Bottom most pixel to be extracted 
        """
    #create function to find location where SNR is the highest
    #loop from the highest value of the signal downward 
        sn_old= 0
        for ii in range(0,cen+1): #0,1,2,3,4,5 range(0,6).. center=5
            sig_sum = sum(signal_col[cen-ii:cen+1])
            noi_sum = np.sqrt(sum(noise_col[cen-ii:cen+1]))
            bkg_sum = sum(bkg_col[cen-ii:cen+1])
            sn_new = sig_sum/noi_sum
            if sn_old >=sn_new:               
                return cen-ii+1                     
            else:
                sn_old = sn_new
        return 0


    def loopingU(self, cen, signal_col, noise_col, bkg_col):
        """Finds top of the optimal extraction region.
    
        Find location where SNR is the highest and loop from highest value upward 
    
        Parameters
        ----------
        cen : float or int 
            Pixel where SNR is the highest 
        signal_col : array of float 
            Array of fluxes to be extracted in a single column on the detector 
        noise_col : array of float 
            Array of noise to be extracted in a single column on the detector 
        bkg_col : array of float 
            Array of background fluxes to be extracted in a single column on the detector 
    
        Return 
        ------
        int
            Top most pixel to be extracted 
        """
        sn_old= 0
        for ii in range(1,(len(signal_col)-cen+1)): #1,2,3,4,5,6 range(1,6).. center=5, edge =10
            sig_sum = sum(signal_col[cen:cen+ii])
            noi_sum = np.sqrt(sum(noise_col[cen:cen+ii]))
            bkg_sum = sum(bkg_col[cen:cen+ii])
            sn_new = sig_sum/noi_sum

            if sn_old >=sn_new:               
                return cen+ii-1                     
            else:
                sn_old = sn_new
        return len(signal_col)+1 


    def sum_spatial(self, extract_info):    
        """Sum pixel in the spatial direction 
        
        Takes extraction info from `extract_region` and sums pixels in that region 
        taking into account integrations and number of transits 
        
        Parameters 
        ----------
        extract_info : dict 
            Dictionary with information on extraction box, flux and noise products
            
        Return
        ------
        dict 
            Dictionary with all extracted 1d products including noise, background, fluxes 
        """
        nint_in = self.nint_in
        nint_out = self.nint_out
        LBout = extract_info['bounds']['LBout']
        LBin = extract_info['bounds']['LBin']
        UBin = extract_info['bounds']['UBin']
        UBout = extract_info['bounds']['UBout']
        photon_sig_in = extract_info['photons']['photon_sig_in']    
        photon_sig_out = extract_info['photons']['photon_sig_out']
        var_pix_in = extract_info['photons']['var_pix_in']
        var_pix_out = extract_info['photons']['var_pix_out']
        photon_sky_in = extract_info['noise']['photon_sky_in']
        photon_sky_out = extract_info['noise']['photon_sky_out']
        rn_var_in = extract_info['noise']['rn_var_in']
        rn_var_out = extract_info['noise']['rn_var_out']


        rr, lenw = photon_sig_in.shape

        # sum spectrum in the spatial direction to create 1d spectrum 
        photon_out_1d = range(0,lenw)
        photon_in_1d = range(0,lenw)
        var_out_1d = range(0,lenw)
        var_in_1d = range(0,lenw)
        sky_in_1d = range(0,lenw)
        sky_out_1d = range(0,lenw)
        rn_in_1d = range(0,lenw)
        rn_out_1d = range(0,lenw)


        #sum 2d spectrum in extraction region in spatial direciton to create 1d spec
        for i in range(0,lenw):
            photon_out_1d[i] = sum(photon_sig_out[LBout[i]:UBout[i],i])*nint_out
            photon_in_1d[i] = sum(photon_sig_in[LBout[i]:UBout[i],i])*nint_in
            var_out_1d[i] = sum(var_pix_out[LBout[i]:UBout[i],i])*nint_out
            var_in_1d[i] = sum(var_pix_in[LBout[i]:UBout[i],i])*nint_in
            sky_out_1d[i] = sum(photon_sky_out[LBout[i]:UBout[i],i])*nint_out
            sky_in_1d[i] = sum(photon_sky_in[LBout[i]:UBout[i],i])*nint_in
            rn_out_1d[i] = sum(rn_var_out[LBout[i]:UBout[i],i])*nint_out
            rn_in_1d[i] = sum(rn_var_in[LBout[i]:UBout[i],i])*nint_in


        photon_out_1d = np.array(photon_out_1d)
        photon_in_1d = np.array(photon_in_1d)
        var_out_1d = np.array(var_out_1d)
        var_in_1d = np.array(var_in_1d)
        sky_in_1d = np.array(sky_in_1d)
        sky_out_1d = np.array(sky_out_1d)
        rn_out_1d = np.array(rn_out_1d)
        rn_in_1d = np.array(rn_in_1d)

        return {'photon_out_1d':photon_out_1d, 'photon_in_1d':photon_in_1d, 
                'var_in_1d':var_in_1d, 'var_out_1d':var_out_1d,
                'rn[out,in]':[rn_out_1d,rn_in_1d],'bkg[out,in]': [sky_in_1d,sky_out_1d], 
                'extract_info':extract_info,'on_source_in':self.on_source_in, 
                'on_source_out':self.on_source_out}

    def extract_region(self): #second to last 
        """Determine extraction Region
        
        Contains functionality to determine extraction region from Pandeia 2d noise 
        simulations. Calls `self.loopingL` and `self.loopingU`. 
        
        Return
        ------
        dict 
            bounding regions, photon and noise to be summed 
        """
        inn = self.inn
        out = self.out
        exptime_per_int = self.exptime_per_int
        ngroups_per_int = self.ngroups_per_int
        #Full variance including all noise sources
        detector_var=out.noise.var_pix
        #Standard deviation (i.e. sqrt(detector_var))
        detector_stdev=out.noise.stdev_pix

        #signal (no noise no background)
        s_out = out.signals[0].rate 
        s_in = inn.signals[0].rate 

        #size of wavelength direction 
        rr,lenw = s_out.shape

        #background 
        bkgd_out = out.signals[0].rate_plus_bg - out.signals[0].rate
        bkgd_in = inn.signals[0].rate_plus_bg - inn.signals[0].rate


        #define psf center based on max value on detector
        cenRo,cenCo = np.unravel_index(s_out.argmax(), s_out.shape)
        cenRi,cenCi = np.unravel_index(s_out.argmax(), s_in.shape)

        #define out of transit parameters to calculate extraction region

        #multiaccum sample data factor 
        factor_flux = 1.0 #6.0/5.0*(ngroups_per_int**2.0+1.0)/(ngroups_per_int**2.0+ngroups_per_int)
        factor_rn = 1.0 #12.0*(ngroups_per_int-1.0)/(ngroups_per_int**2.0 + ngroups_per_int)

        photon_sig_out = s_out*exptime_per_int*factor_flux #total photons per pixel in signal
        photon_sky_out = bkgd_out*exptime_per_int*factor_flux #total photons per pixel in background 
        #variance stricly due to detector readnoise.. You might think this is 
        #wrong because usually RN isnt multiplie by time.. but Pandeia gives RN in rms/ sec. 
        rn_var_out= out.noise.var_rn_pix*exptime_per_int*factor_rn 

        var_pix_out = photon_sig_out + photon_sky_out + rn_var_out # variance of noise per pixel

        #define parameters for IN transit 
        photon_sig_in = s_in*exptime_per_int*factor_flux #total photons per pixel in signal
        photon_sky_in = bkgd_in*exptime_per_int*factor_flux #total photons per pixel in background 
        rn_var_in= inn.noise.var_rn_pix*exptime_per_int*factor_rn #variance stricly due to detector readnoise
        var_pix_in = photon_sig_in + photon_sky_in + rn_var_in # variance of noise per pixel 

        UBout = range(0,lenw)
        LBout = range(0,lenw)
        UBin = range(0,lenw)
        LBin = range(0,lenw)


        #start new loop over column
        for j in range(0,lenw):

            #define column of interest
            noise_col_out = var_pix_out[:, j]
            signal_col_out = photon_sig_out[:,j]
            bkg_col_out = photon_sky_out[:,j]
            rn_col_out = rn_var_out[:,j]

            noise_col_in =  var_pix_in[:, j]
            signal_col_in = photon_sig_in[:,j]
            bkg_col_in = photon_sky_in[:,j]


            #store lower and upper bound extraction regions for in and out of transit 
            LBout[j] = self.loopingL(cenRo, signal_col_out, noise_col_out, bkg_col_out)
            UBout[j] = self.loopingU(cenRo, signal_col_out, noise_col_out, bkg_col_out)
            LBin[j] = self.loopingL(cenRi, signal_col_in, noise_col_in, bkg_col_in    )
            UBin[j] = self.loopingU(cenRi, signal_col_in, noise_col_in, bkg_col_in)

        #this could be made more elegant later... not very efficient 
        noise = {'rn_var_out':rn_var_out, 'rn_var_in':rn_var_in, 
                 'photon_sky_in':photon_sky_in, 'photon_sky_out': photon_sky_out }
        bounds = {'LBout':LBout, 'UBout':UBout, 'LBin':LBin, 'UBin':UBin}
        photons = {'photon_sig_out': photon_sig_out, 'photon_sig_in':photon_sig_in, 
                   'var_pix_in':var_pix_in,'var_pix_out': var_pix_out}
        extract_info ={'bounds':bounds, 'photons':photons, 'noise':noise}

        return extract_info

    def run_2d_extract(self):
        """Extract noise from 2d detector image
            
        Contains functionality to extract noise from 2d detector image
            
        Returns
        -------
        dict 
            all optimally extracted 1d products  
        """
        #optimize SNR and extract region 
        extract = self.extract_region()
        #return summed up pixels 
        return self.sum_spatial(extract)
        
        
    def run_slope_method(self): 
        """Compute noise using Pandeia 1d noise
        
        Contains functionality to compute noise using Pandeia 1d noise 
        output products (uses MULTIACCUM) noise formula
        
        Return
        ------
        dict 
            all optimally extracted 1d products for a single occultation
        """
        #on source out versus in 
        on_source_in = self.on_source_in
        on_source_out = self.on_source_out
        
        curves_out = self.out['1d']
        curves_inn = self.inn['1d']

        #background + contamination extracted 
        #not used for noise calculations, just 
        #used for output
        bkg_flux_inn = curves_inn['extracted_bg_only'][1] * on_source_in 
        bkg_flux_out = curves_out['extracted_bg_only'][1] * on_source_out

        #calculate rn 
        rn_var = 2.0*self.rn**2.0
                
        #1d rn     = rn/pix * # of integrations * #extraction area 
        #not used for noise calcualtions here 
        #just used for output 
        rn_var_inn = rn_var * self.nint_in * self.extraction_area 
        rn_var_out = rn_var * self.nint_out * self.extraction_area 

        #In the following the SN is changed to incorporate number of occultations 
        #i.e. multiply by sqrt(n) 
        sn_in = curves_inn['sn'][1]
        sn_out = curves_out['sn'][1]

        extracted_flux_inn = curves_inn['extracted_flux'][1]*on_source_in
        
        extracted_noise_inn = curves_inn['extracted_flux'][1]/(sn_in)

        extracted_flux_out = curves_out['extracted_flux'][1]*on_source_out
        
        extracted_noise_out = curves_out['extracted_flux'][1]/(sn_out)

        #units of this unconventional.. sigma/s
        #because snr = extracted flux / extracted noise and 
        #extracted flux in units of electrons /s
        varin = (extracted_noise_inn*on_source_in)**2.0
        varout = (extracted_noise_out*on_source_out)**2.0

        return {'photon_out_1d':extracted_flux_out, 'photon_in_1d':extracted_flux_inn, 
                    'var_in_1d':varin, 'var_out_1d': varout,'on_source_in':self.on_source_in, 
                'on_source_out':self.on_source_out,'bkg[out,in]':[bkg_flux_out,bkg_flux_inn],
                'rn[out,in]':[rn_var_out,rn_var_inn]}
    
    
    def run_f_minus_l(self):
        """Compute noise using first minus last formula
        
        Uses 1d exracted products from pandeia to compute noise without 
        multiaccum noise formula from Rauscher 07. Includes readnoise 
        background. 
        
        Returns
        -------
        dict 
            all optimally extracted 1d products for a single transit     
        """

        inn = self.inn
        curves_out = self.out['1d']
        curves_inn = self.inn['1d']
        
        
        #on source out versus in 
        on_source_in = self.on_source_in
        on_source_out = self.on_source_out

        #calculate rn 
        rn_var = self.rn**2.0
                
        #1d rn     = rn/pix * # of integrations   * #pixs
        rn_var_inn = rn_var * self.nint_in * self.extraction_area 
        rn_var_out = rn_var * self.nint_out * self.extraction_area 

        #extract fluxs
        extracted_flux_inn = curves_inn['extracted_flux'][1] * on_source_in 
        extracted_flux_out = curves_out['extracted_flux'][1] * on_source_out 
                
        #background + contamination extracted 
        bkg_flux_inn = curves_inn['extracted_bg_only'][1] * on_source_in 
        bkg_flux_out = curves_out['extracted_bg_only'][1] * on_source_out 
        
        #total nois 
        varin = (extracted_flux_inn + bkg_flux_inn + rn_var_inn)
        varout = (extracted_flux_out + bkg_flux_out + rn_var_out)
        
        return {'photon_out_1d':extracted_flux_out, 'photon_in_1d':extracted_flux_inn, 
                    'var_in_1d':varin, 'var_out_1d': varout,'on_source_in':self.on_source_in, 
                'on_source_out':self.on_source_out, 'rn[out,in]':[rn_var_out,rn_var_inn], 
                'bkg[out,in]':[bkg_flux_out,bkg_flux_inn]}
    

    
    def run_phase_spec(self):
        """Computes time dependent noise for phase curves.
        
        Computes noise for phase curve analysis instead of spectroscopy. 
        Using MULTIACCUM formula here but this could be changed in the future.  Does not 
        return a spectra for each time element... On your own for that. 
        
        Return
        ------
        dict 
            all optimally extracted 1d products        
        """
        time = self.inn['time']
        tint = self.exptime_per_int 
        curves_out = self.out['1d']
        #don't input a ridiculously low res phase curve
        new_t = np.arange(min(time), max(time), tint) 
        planet_phase = np.interp(new_t, time, self.inn['planet_phase'])

        rn_var = 2.0*self.rn**2.0


        #1d rn     = rn/pix * # of integrations * #pixs 
        rn_var_out = rn_var * self.extraction_area  # the initial output is always sampled by 1 integration
   
        extracted_flux_out = sum(curves_out['extracted_flux'][1]) * tint 
        
        flux_time_out = extracted_flux_out+ np.zeros(len(new_t))
        
        flux_time_in = flux_time_out*(1.0 + planet_phase)
        
        bkg_flux_out = sum(curves_out['extracted_bg_only'][1]) * tint 
        
        varout = flux_time_out + bkg_flux_out + rn_var_out
        varin = flux_time_in + bkg_flux_out + rn_var_out
 
        
        return {'photon_out_1d':flux_time_out, 'photon_in_1d':flux_time_in, 
                    'var_in_1d':varin, 'var_out_1d': varout, 'time':new_t,'on_source_in':'N/A', 
                'on_source_out':'N/A','rn[out,in]':[rn_var_out,rn_var_out], 
                'bkg[out,in]':[bkg_flux_out,bkg_flux_out]}
        




import json
import os
import uuid
from collections import namedtuple, OrderedDict
import tornado.escape
import tornado.httpserver
import tornado.ioloop
import tornado.options
import tornado.web
from tornado.options import define, options

import traceback
from sqlalchemy import *
from concurrent.futures import ProcessPoolExecutor
import pickle
import pandas as pd 
import numpy as np
import requests
from astroquery.simbad import Simbad
import astropy.units as u

from .pandexo import wrapper
from .utils.plotters import create_component_jwst, create_component_hst
from .logs import jwst_log, hst_log
from .exomast import get_target_data


# define location of temp files
__TEMP__ = os.environ.get("PANDEXO_TEMP", os.path.join(os.path.dirname(__file__), "temp"))

#define location of fort grids
try:
    __FORT__ = os.environ.get('FORTGRID_DIR')
    db_fort = create_engine('sqlite:///'+__FORT__)
except: 
    print('FORTNEY DATABASE NOT INSTALLED')

#add Simbad query info 
Simbad.add_votable_fields('flux(H)')
Simbad.add_votable_fields('flux(J)')

define("port", default=1111, help="run on the given port", type=int)
define("debug", default=False, help="automatically detect code changes in development")
define("workers", default=4, help="maximum number of simultaneous async tasks")

# Define a simple named tuple to keep track for submitted calculations
CalculationTask = namedtuple('CalculationTask', ['id', 'name', 'task',
                                                 'cookie', 'count'])


class Application(tornado.web.Application):
    """Gobal settings of the server
    This defines the global settings of the server. This parses out the
    handlers, and includes settings for if ever we want to tie this to a
    database.
    """
    def __init__(self):
        handlers = [
            (r"/", HomeHandler),
            (r"/about", AboutHandler),
            (r"/dashboard", DashboardHandler),
            (r"/dashboardhst", DashboardHSTHandler),
            (r"/tables", TablesHandler),
            (r"/helpfulplots", HelpfulPlotsHandler),
            (r"/calculation/new", CalculationNewHandler),
            (r"/calculation/newHST", CalculationNewHSTHandler),
            (r"/calculation/status/([^/]+)", CalculationStatusHandler),
            (r"/calculation/statushst/([^/]+)", CalculationStatusHSTHandler),
            (r"/calculation/view/([^/]+)", CalculationViewHandler),
            (r"/calculation/viewhst/([^/]+)", CalculationViewHSTHandler),
            (r"/calculation/download/([^/]+)", CalculationDownloadHandler),
            (r"/calculation/downloadpandin/([^/]+)", CalculationDownloadPandInHandler)
        ]
        settings = dict(
            blog_title="Pandexo",
            template_path=os.path.join(os.path.dirname(__file__), "templates"),
            static_path=os.path.join(os.path.dirname(__file__), "static"),
            xsrf_cookies=True,
            cookie_secret="__TODO:_GENERATE_YOUR_OWN_RANDOM_VALUE_HERE__",
            debug=True,
        )
        super(Application, self).__init__(handlers, **settings)


class BaseHandler(tornado.web.RequestHandler):
    """
    Logic to handle user information and database access might go here.
    """
    executor = ProcessPoolExecutor(max_workers=16)
    buffer = OrderedDict()

    def _get_task_response(self, id):
        """
        Simple function to grab a calculation that's stored in the buffer,
        and return a dictionary/json-like response to the front-end.
        """
        calc_task = self.buffer.get(id)
        task = calc_task.task

        response = {'id': id,
                    'name': calc_task.name,
                    'count': calc_task.count}

        if task.running():
            response['state'] = 'running'
            response['code'] = 202
        elif task.done():
            response['state'] = 'finished'
        elif task.cancelled():
            response['state'] = 'cancelled'
        else:
            response['state'] = 'pending'

        response['html'] = tornado.escape.to_basestring(
            self.render_string("calc_row.html", response=response))
        return response

 
    def _get_task_response_hst(self, id):
        """
        Simple function to grab a calculation that's stored in the buffer,
        and return a dictionary/json-like response to the front-end.
        """
        calc_task = self.buffer.get(id)
        task = calc_task.task

        response = {'id': id,
                    'name': calc_task.name,
                    'count': calc_task.count}

        if task.running():
            response['state'] = 'running'
            response['code'] = 202
        elif task.done():
            response['state'] = 'finished'
        elif task.cancelled():
            response['state'] = 'cancelled'
        else:
            response['state'] = 'pending'

        response['html'] = tornado.escape.to_basestring(
            self.render_string("calc_rowhst.html", response=response))

        return response
        
    def write_error(self, status_code, **kwargs):
        """
        This renders a customized error page
        """
        reason = self._reason
        error_info = ''
        trace_print = ''
        if 'exc_info' in kwargs:
            error_info = kwargs['exc_info']
            try:
                trace_print = traceback.format_exception(*error_info)
                trace_print = "\n".join(map(str,trace_print))
            except:
                pass
        self.render('errors.html',page=None, status_code=status_code, reason=reason, error_log=trace_print)


    def _get_task_result(self, id):
        """
        This method grabs only the result returned from the python `Future`
        object. This contains the stuff that Pandeia returns.
        """
        calc_task = self.buffer.get(id)
        task = calc_task.task

        return task.result()

    def _add_task(self, id, name, task):
        """
        This creates the task and adds it to the buffer.
        """
        self.buffer[id] = CalculationTask(id=id, name=name, task=task,
                                          count=len(self.buffer)+1,
                                          cookie=self.get_cookie("pandexo_user"))

        # Only allow 100 tasks **globally**. This will delete old tasks first.
        if len(self.buffer) > 100:
            self.buffer.popitem(last=False)


class HomeHandler(BaseHandler):
    def get(self):
        """
        This sets an **unsecured** cookie. If user accounts gets
        implemented, this must be changed to a secure cookie.
        """
        if not self.get_cookie("pandexo_user"):
            self.set_cookie("pandexo_user", str(uuid.uuid4()))

        self.render("home.html")
        
class AboutHandler(BaseHandler):
    def get(self):
        """
        Render about PandExo Page
        """
        self.render("about.html")

class TablesHandler(BaseHandler):
    def get(self):
        """
        Render tables with confirmed candidates
        """
        self.render("tables.html")

class HelpfulPlotsHandler(BaseHandler):
    def get(self):
        """
        Renders helpful bokeh plots
        """
        self.render("helpfulplots.html")


class DashboardHandler(BaseHandler):
    """
    Request handler for the dashboard page. This will retrieve and render
    the html template, along with the list of current task objects.
    """
    def get(self):
        task_responses = [self._get_task_response(id) for id, nt in
                          list(self.buffer.items())
                          if ((nt.cookie == self.get_cookie("pandexo_user"))
                          & (id[len(id)-1]=='e'))]
        
        self.render("dashboard.html", calculations=task_responses[::-1])


class DashboardHSTHandler(BaseHandler):
    """
    Request handler for the dashboard page. This will retrieve and render
    the html template, along with the list of current task objects.
    """
    def get(self):
        task_responses = [self._get_task_response_hst(id) for id, nt in
                          list(self.buffer.items())
                          if ((nt.cookie == self.get_cookie("pandexo_user"))
                          & (id[len(id)-1]=='h'))]
        
        self.render("dashboardhst.html", calculations=task_responses[::-1])


class CalculationNewHandler(BaseHandler):
    """
    This request handler deals with processing the form data and submitting
    a new calculation task to the parallelized workers.
    """
    def get(self):
        try: 
            header= pd.read_sql_table('header',db_fort)
        except:
            header = pd.DataFrame({
            'temp': ['NO GRID DB FOUND'],
            'ray' : ['NO GRID DB FOUND'],
            'flat':['NO GRID DB FOUND']})

        with open(os.path.join(os.path.dirname(__file__), "reference",
                               "exo_input.json")) as data_file:
            exodata = json.load(data_file)

        all_planets =  requests.get("https://exoplanetarchive.ipac.caltech.edu/cgi-bin/nstedAPI/nph-nstedAPI?table=exoplanets&select=pl_name&format=csv")  
        all_planets = all_planets.text.replace(' ','').split('\n')[1:]

        self.render("new.html", id=id,
                                 temp=list(map(str, header.temp.unique())), 
                                 planets=all_planets,
                                 data=exodata)

    def post(self):
        """
        The post method contains the returned data from the form data (
        accessed by using `self.get_argument(...)` for specific arguments,
        or `self.request.body` to grab the entire returned object.
        """
        
        #print(self.request.body)
        
        id = str(uuid.uuid4())+'e'
       
        # Read data from exoMAST, print it on the webpage:
        exomast_response = self.get_argument("resolve_target", None)
        submit_form = self.get_argument("submit_form", None)

        if exomast_response=="exomast":
            with open(os.path.join(os.path.dirname(__file__), "reference",
                            "exo_input.json")) as data_file:
                exodata = json.load(data_file)
                exodata["telescope"] = 'jwst'
                exodata["calculation"] = 'fml'  # always for online form

            try:
                planet_name = self.get_argument("planetname")
                planet_data = get_target_data(planet_name)[0]
                exodata['planet']['planetname'] = planet_data['canonical_name']
                # for item in planet_data:
                #     print("{}: {}".format(item, planet_data[item]))
                # star
                exodata["star"]["temp"] = planet_data['Teff']
                exodata["star"]["logg"] = planet_data['stellar_gravity']
                exodata["star"]["metal"] = planet_data['Fe/H'] 
                # Keep Simbad query, as exoMAST typically does not have Jmag:
                exodata["star"]["jmag"] = Simbad.query_object(planet_name[:-1])['FLUX_J'][0] #planet_data['Jmag']
                exodata["star"]["ref_wave"] = 1.25

                # optional star radius
                exodata["star"]["radius"] = planet_data['Rs']  
                exodata["star"]["r_unit"] = planet_data['Rs_unit'][0]+ planet_data['Rs_unit'][1:].lower()    
 
                # optional planet radius/mass
                exodata["planet"]["radius"] = planet_data['Rp']  
                exodata["planet"]["r_unit"] = planet_data['Rp_unit'][0]+ planet_data['Rp_unit'][1:].lower()  
                exodata["planet"]["mass"] = planet_data['Mp'] 
                exodata["planet"]["m_unit"] = planet_data['Mp_unit'][0]+ planet_data['Mp_unit'][1:].lower() 
                exodata["planet"]["transit_duration"] = planet_data['transit_duration']
                exodata["planet"]["td_unit"] = planet_data['transit_duration_unit']

                if planet_data['inclination'] == None:   
                    inc = 90
                else: 
                    inc = planet_data['inclination']

                exodata["planet"]["i"]          = inc
                exodata["planet"]["ars"]        = planet_data['a/Rs'] 
                period = planet_data['orbital_period'] 
                period_unit = planet_data['orbital_period_unit'] 
                exodata["planet"]["period"]     = (period*u.Unit(period_unit)).to(u.Unit('day')).value
                exodata["planet"]["ecc"]        = planet_data['eccentricity'] 
                try:
                    exodata["planet"]["w"]      = float(planet_data['omega'] )
                except: 
                    exodata["planet"]["w"]      = 90.

                self.header = pd.DataFrame({'temp': ['NO GRID DB FOUND'],
                                    'ray' : ['NO GRID DB FOUND'],
                                    'flat':['NO GRID DB FOUND']})
            except:
                exodata['url_err'] = 'Sorry, cant resolve target {}'.format(planet_name)

            all_planets =  requests.get("https://exoplanetarchive.ipac.caltech.edu/cgi-bin/nstedAPI/nph-nstedAPI?table=exoplanets&select=pl_name&format=csv")  
            all_planets = all_planets.text.replace(' ','').split('\n')[1:]

            self.header = pd.DataFrame({'temp': ['NO GRID DB FOUND'],
                                    'ray' : ['NO GRID DB FOUND'],
                                    'flat':['NO GRID DB FOUND']})

            return self.render("new.html", id=id,
                                temp=list(map(str, self.header.temp.unique())),
                                data=exodata,
                                planets=all_planets)


        if submit_form == 'submit':
            # If submit, read all submitted properties. This is the same as if properties=="user", 
            # because either exoMAST was successful and populated everything, the user changed some of 
            # those properties *or* exoMAST was not succesful and user had to fill everything:

            with open(os.path.join(os.path.dirname(__file__), "reference",
                            "exo_input.json")) as data_file:

                exodata = json.load(data_file)
                exodata["telescope"] = 'jwst'
                exodata["calculation"] = 'fml'

                #star

                exodata["star"]["temp"] = float(self.get_argument("temp"))
                exodata["star"]["logg"] = float(self.get_argument("logg"))
                exodata["star"]["metal"] = float(self.get_argument("metal"))  
                exodata["star"]["mag"] = float(self.get_argument("mag"))
                exodata["star"]["ref_wave"] = float(self.get_argument("ref_wave"))

                #optinoal star radius
                exodata["star"]["radius"] = float(self.get_argument("rstarc"))
                exodata["star"]["r_unit"] = str(self.get_argument("rstar_unitc"))
 
                #optional planet radius
                exodata["planet"]["radius"] = float(self.get_argument("refradc"))
                exodata["planet"]["r_unit"] = str(self.get_argument("r_unitc")) 

                #transit duration
                # for phase curves user doesn't necessarily have to input a transit duration
                try:
                    exodata["planet"]["transit_duration"] = float(self.get_argument("transit_duration"))
                    exodata["planet"]["td_unit"] = str(self.get_argument("td_unit"))
                except:
                    # but if they don't.. make sure that the planet units are in seconds...
                    if self.get_argument("planwunits") == 'sec':
                        exodata["planet"]["transit_duration"] = 0.0
                    else: 
                        raise Exception("Need transit duraiton or input phase curve file")

                # stellar model
                exodata["star"]["type"] = self.get_argument("stellarModel")

                if exodata["star"]["type"] == "user":
                    # process star file
                    fileinfo_star = self.request.files['starFile'][0]
                    fname_star = fileinfo_star['filename']
                    extn_star = os.path.splitext(fname_star)[1]
                    cname_star = id+'star' + extn_star
                    fh_star = open(os.path.join(__TEMP__, cname_star), 'wb')
                    fh_star.write(fileinfo_star['body'])

                    exodata["star"]["starpath"] = os.path.join(__TEMP__, cname_star)
                    exodata["star"]["f_unit"] = self.get_argument("starfunits")
                    exodata["star"]["w_unit"] = self.get_argument("starwunits")
     

                # planet model
                exodata["planet"]["type"] = self.get_argument("planetModel")
                if exodata["planet"]["type"] == "user":
                    # process planet file
                    fileinfo_plan = self.request.files['planFile'][0]
                    fname_plan = fileinfo_plan['filename']
                    extn_plan = os.path.splitext(fname_plan)[1]
                    cname_plan = id+'planet' + extn_plan
                    fh_plan = open(os.path.join(__TEMP__, cname_plan), 'wb')
                    fh_plan.write(fileinfo_plan['body'])

                    exodata["planet"]["exopath"] = os.path.join(__TEMP__, cname_plan)
                    exodata["planet"]["w_unit"] = self.get_argument("planwunits")
                    exodata["planet"]["f_unit"] = self.get_argument("planfunits")
                elif exodata["planet"]["type"] == "constant":                               
                    if self.get_argument("constant_unit") == 'fp/f*':
                        exodata["planet"]["temp"] = float(self.get_argument("ptempc"))
                        exodata["planet"]["f_unit"] = 'fp/f*'
                    elif self.get_argument("constant_unit") == 'rp^2/r*^2':
                        exodata["planet"]["f_unit"] = 'rp^2/r*^2'
                elif exodata["planet"]["type"] == "grid":
                    exodata["planet"]["temp"] = float(self.get_argument("ptempg"))
                    exodata["planet"]["chem"] = str(self.get_argument("pchem"))
                    exodata["planet"]["cloud"] = self.get_argument("cloud") 
                    exodata["planet"]["mass"] = float(self.get_argument("pmass"))
                    exodata["planet"]["m_unit"] = str(self.get_argument("m_unit"))
                #baseline 
                exodata["observation"]["baseline"] = float(self.get_argument("baseline"))
                exodata["observation"]["baseline_unit"] = self.get_argument("baseline_unit")
                try:
                    exodata["observation"]["target_acq"] = self.get_argument("TA") == 'on'
                except:
                    exodata["observation"]["target_acq"] = False
                    
                exodata["observation"]["noccultations"] = float(self.get_argument("numtrans"))
                exodata["observation"]["sat_level"] = float(self.get_argument("satlevel"))
                exodata["observation"]["sat_unit"] = self.get_argument("sat_unit")

                    
                # noise floor, set to 0.0 of no values are input
                try:
                    observation_type = self.get_argument("noiseModel")
                    if observation_type == "user":
                        # process noise file
                        fileinfo_noise = self.request.files['noiseFile'][0]
                        fname_noise = fileinfo_noise['filename']
                        extn_noise = os.path.splitext(fname_noise)[1]
                        cname_noise = id + 'noise' + extn_noise
                        fh_noise = open(os.path.join(__TEMP__, cname_noise), 'wb')
                        fh_noise.write(fileinfo_star['body'])
                        exodata["observation"]["noise_floor"] = os.path.join(__TEMP__, cname_noise)
                    else:
                        exodata["observation"]["noise_floor"] = float(self.get_argument("noisefloor"))
                except:
                    exodata["observation"]["noise_floor"] = 0.0

            instrument = self.get_argument("instrument").lower()
            if instrument == "miri":
                with open(os.path.join(os.path.dirname(__file__), "reference", "miri_input.json")) as data_file:
                    pandata = json.load(data_file)       
                    mirimode = self.get_argument("mirimode")
                    if (mirimode == "lrsslit"):
                        pandata["configuration"]["instrument"]["mode"] = mirimode
                        pandata["configuration"]["instrument"]["aperture"] = "lrsslit"
                        pandata["configuration"]["detector"]["subarray"] = "full"

            if instrument == "nirspec":
                with open(os.path.join(os.path.dirname(__file__), "reference", "nirspec_input.json")) as data_file:
                    pandata = json.load(data_file)  
                    nirspecmode = self.get_argument("nirspecmode")
                    pandata["configuration"]["instrument"]["disperser"] = nirspecmode[0:5]
                    pandata["configuration"]["instrument"]["filter"] = nirspecmode[5:11]
                    pandata["configuration"]["detector"]["subarray"] = self.get_argument("nirspecsubarray")

            if instrument == "nircam":
                with open(os.path.join(os.path.dirname(__file__), "reference", "nircam_input.json")) as data_file:
                    pandata = json.load(data_file) 
                    pandata["configuration"]["instrument"]["filter"] = self.get_argument("nircammode")
                    pandata["configuration"]["detector"]["subarray"] = self.get_argument("nircamsubarray")

            if instrument == "niriss":
                with open(os.path.join(os.path.dirname(__file__), "reference", "niriss_input.json")) as data_file:
                    pandata = json.load(data_file)
                    nirissmode = self.get_argument("nirissmode")
                    pandata["configuration"]["detector"]["subarray"] = nirissmode

            pandata['configuration']['instrument']['instrument'] = instrument
            
            # write in optimal groups or set a number
            try:
                pandata["configuration"]["detector"]["ngroup"] = int(self.get_argument("optimize"))
            except: 
                pandata["configuration"]["detector"]["ngroup"] = self.get_argument("optimize")
            
            finaldata = {"pandeia_input": pandata, "pandexo_input": exodata}

            #PandExo stats
            try: 
                jwst_log(finaldata)
            except: 
                pass

            task = self.executor.submit(wrapper, finaldata)


            self._add_task(id, self.get_argument("calcName"), task)

            response = self._get_task_response(id)
            response['info'] = {}
            response['location'] = '/calculation/status/{}'.format(id)
            
            
            self.write(dict(response))
            self.redirect("../dashboard")

class CalculationNewHSTHandler(BaseHandler):
    """
    This request handler deals with processing the form data and submitting
    a new HST calculation task to the parallelized workers.
    """


    def get(self):
        try: 
            self.header= pd.read_sql_table('header',db_fort)
        except:
            self.header = pd.DataFrame({
            'temp': ['NO GRID DB FOUND'],
            'ray' : ['NO GRID DB FOUND'],
            'flat':['NO GRID DB FOUND']})
        with open(os.path.join(os.path.dirname(__file__), "reference",
                               "exo_input.json")) as data_file:
            exodata = json.load(data_file)
        all_planets =  requests.get("https://exoplanetarchive.ipac.caltech.edu/cgi-bin/nstedAPI/nph-nstedAPI?table=exoplanets&select=pl_name&format=csv")  
        all_planets = all_planets.text.replace(' ','').split('\n')[1:]
        self.render("newHST.html", id=id,
                    temp=list(map(str, self.header.temp.unique())),
                    data=exodata,
                    planets=all_planets)

    def post(self):
        """
        The post method contains the retured data from the form data (
        accessed by using `self.get_argument(...)` for specific arguments,
        or `self.request.body` to grab the entire returned object.
        """
        
        # print(self.request.body)

        id = str(uuid.uuid4())+'h'

        exomast_response = self.get_argument("resolve_target", None)
        
        submit_form = self.get_argument("submit_form", None)

        if exomast_response=="exomast":
            with open(os.path.join(os.path.dirname(__file__), "reference",
                            "exo_input.json")) as data_file:
                exodata = json.load(data_file)
                exodata["telescope"] = 'hst'

            try:
                planet_name = self.get_argument("planetname")
                planet_data, url = get_target_data(planet_name)

                exodata['planet']['planetname'] = planet_data['canonical_name']
                exodata['url'] = url
                
                #star
                exodata["star"]["temp"] = planet_data['Teff']

                jmag = Simbad.query_object(planet_name[:-1])['FLUX_J'][0]
                hmag = Simbad.query_object(planet_name[:-1])['FLUX_H'][0]

                exodata["star"]["jmag"] = jmag
                exodata["star"]["hmag"] = hmag

                if exodata["telescope"] == 'hst':
                    exodata["star"]["mag"] = hmag

                #optinoal star radius
                exodata["star"]["radius"] = planet_data['Rs']  
                exodata["star"]["r_unit"] = planet_data['Rs_unit'][0]+ planet_data['Rs_unit'][1:].lower()    

                #optional planet radius/mass
                exodata["planet"]["radius"] = planet_data['Rp']
                exodata["planet"]["r_unit"] = planet_data['Rp_unit'][0]+ planet_data['Rp_unit'][1:].lower()  
                exodata["planet"]["mass"] = planet_data['Mp'] 
                exodata["planet"]["m_unit"] = planet_data['Mp_unit'][0]+ planet_data['Mp_unit'][1:].lower()  

                transit_duration = planet_data['transit_duration'] 
                td_unit = planet_data['transit_duration_unit'] 
                transit_duration  = (transit_duration*u.Unit(td_unit)).to(u.Unit('day')).value
                exodata["planet"]["transit_duration"] = transit_duration
                
                if planet_data['inclination'] == None:   
                    inc = 90
                else: 
                    inc = planet_data['inclination']

                exodata["planet"]["i"]          = inc
                exodata["planet"]["ars"]        = planet_data['a/Rs'] 
                period = planet_data['orbital_period'] 
                period_unit = planet_data['orbital_period_unit'] 
                exodata["planet"]["period"]     = (period*u.Unit(period_unit)).to(u.Unit('day')).value
                exodata["planet"]["ecc"]        = planet_data['eccentricity'] 
                try:
                    exodata["planet"]["w"]      = float(planet_data['omega'] )
                except: 
                    exodata["planet"]["w"]      = 90.

                self.header = pd.DataFrame({'temp': ['NO GRID DB FOUND'],
                                    'ray' : ['NO GRID DB FOUND'],
                                    'flat':['NO GRID DB FOUND']})
            except:
                exodata['url_err'] = 'Sorry, cant resolve target {}'.format(planet_name)

            all_planets =  requests.get("https://exoplanetarchive.ipac.caltech.edu/cgi-bin/nstedAPI/nph-nstedAPI?table=exoplanets&select=pl_name&format=csv")  
            all_planets = all_planets.text.replace(' ','').split('\n')[1:]

            return self.render("newHST.html", id=id,
                                temp=list(map(str, self.header.temp.unique())),
                                data=exodata,
                                planets=all_planets)

        if submit_form == 'submit':
            with open(os.path.join(os.path.dirname(__file__), "reference",
                            "exo_input.json")) as data_file:
                exodata = json.load(data_file)
                exodata["telescope"] = 'hst'

                #star
                exodata["star"]["jmag"]         = float(self.get_argument("Jmag"))
                try:
                    #only needed for higher accuracy
                    exodata["star"]["hmag"]     = float(self.get_argument("Hmag"))
                except:
                    exodata["star"]["hmag"]     = None

                exodata["star"]["radius"] = float(self.get_argument("rstarc"))
                exodata["star"]["r_unit"] = str(self.get_argument("rstar_unitc"))
                try:
                    #only needed for secondary eclipse
                    exodata["star"]["temp"] = float(self.get_argument("stempc"))
                except:
                    exodata["star"]["temp"] = None

                #planet
                exodata["planet"]["radius"] = float(self.get_argument("refradc"))
                exodata["planet"]["r_unit"] = str(self.get_argument("r_unitc"))
                depth = exodata["planet"]["radius"]**2 / ((exodata["star"]["radius"]
                                                            *u.Unit(exodata["star"]["r_unit"]) )
                                                                .to(u.Unit(exodata["planet"]["r_unit"]))).value**2

                exodata["planet"]["depth"]      = depth
                exodata["planet"]["i"]          = float(self.get_argument("i"))
                exodata["planet"]["ars"]        = float(self.get_argument("ars"))
                exodata["planet"]["period"]     = float(self.get_argument("period"))
                exodata["planet"]["ecc"]        = float(self.get_argument("ecc"))
                try:
                    exodata["planet"]["w"]      = float(self.get_argument("w"))
                except:
                    exodata["planet"]["w"]      = 90.
                exodata["planet"]["transit_duration"]   = float(self.get_argument("transit_duration"))

                # planet model
                exodata["planet"]["type"] = self.get_argument("planetModel")

                if exodata["planet"]["type"] == "user":
                    # process planet file
                    fileinfo_plan = self.request.files['planFile'][0]
                    fname_plan = fileinfo_plan['filename']
                    extn_plan = os.path.splitext(fname_plan)[1]
                    cname_plan = id+'planet' + extn_plan
                    fh_plan = open(os.path.join(__TEMP__, cname_plan), 'wb')
                    fh_plan.write(fileinfo_plan['body'])

                    exodata["planet"]["exopath"] = os.path.join(__TEMP__, cname_plan)
                    exodata["planet"]["w_unit"] = self.get_argument("planwunits")
                    exodata["planet"]["f_unit"] = self.get_argument("planfunits")

                elif exodata["planet"]["type"] == "constant":                               
                    if self.get_argument("constant_unit") == 'fp/f*':
                        exodata["planet"]["temp"] = float(self.get_argument("ptempc"))
                        exodata["planet"]["f_unit"] = 'fp/f*'
                    elif self.get_argument("constant_unit") == 'rp^2/r*^2':
                        exodata["planet"]["f_unit"] = 'rp^2/r*^2'

                elif exodata["planet"]["type"] == "grid":
                    exodata["planet"]["mass"] = float(self.get_argument("pmass"))
                    exodata["planet"]["m_unit"] = str(self.get_argument("m_unit"))
                    exodata["planet"]["temp"] = float(self.get_argument("ptempg"))
                    exodata["planet"]["chem"] = str(self.get_argument("pchem"))
                    exodata["planet"]["cloud"] = self.get_argument("cloud") 

                exodata["observation"]["noise_floor"]           = 0.0
                exodata["calculation"]                          = 'scale'
                    
            if (self.get_argument("instrument")=="STIS"): 
                with open(os.path.join(os.path.dirname(__file__), "reference",
                                   "stis_input.json")) as data_file:   
                    pandata = json.load(data_file)       
                    stismode = self.get_argument("stismode")
            if (self.get_argument("instrument")=="WFC3"): 
                with open(os.path.join(os.path.dirname(__file__), "reference",
                                   "wfc3_input.json")) as data_file:
                    pandata = json.load(data_file)  
                    pandata["configuration"]['detector']['subarray']    = self.get_argument("subarray")
                    pandata["configuration"]['detector']['nsamp']       = int(self.get_argument("nsamp"))
                    pandata["configuration"]['detector']['samp_seq']    = self.get_argument("samp_seq")
                    pandata["configuration"]['instrument']['disperser'] = self.get_argument("wfc3mode")
                try: 
                    pandata["strategy"]["norbits"]           = int(self.get_argument("norbits"))
                except:
                    pandata["strategy"]["norbits"]           = None
                exodata["observation"]["noccultations"]         = int(self.get_argument("noccultations"))
                pandata["strategy"]["nchan"]                 = int(self.get_argument("nchan"))
                pandata["strategy"]["scanDirection"]         = self.get_argument("scanDirection")
                pandata["strategy"]["useFirstOrbit"]         = self.get_argument("useFirstOrbit").lower() == 'true'
                try:
                    pandata["strategy"]["windowSize"]        = float(self.get_argument("windowSize"))
                except:
                    pandata["strategy"]["windowSize"]        = 20.
                pandata["strategy"]["schedulability"]           = self.get_argument("schedulability")
            try:
                calc_ramp = self.get_argument("ramp")

                calc_ramp = True
            except: 
                calc_ramp = False


            pandata['strategy']['calculateRamp'] = calc_ramp
            pandata['strategy']['targetFluence'] = float(self.get_argument("targetFluence"))
                    
            #import pickle as pk 
            #a = pk.load(open('/Users/natashabatalha/Desktop/JWST/testing/ui.pk','rb'))
            #pandata = a['pandeia_input']
            #exodata  = a['pandexo_input']

            finaldata = {"pandeia_input": pandata , "pandexo_input":exodata}
            #PandExo stats
            try: 
                hst_log(finaldata)
            except: 
                pass

            task = self.executor.submit(wrapper, finaldata)

            self._add_task(id, self.get_argument("calcName"), task)

            response = self._get_task_response_hst(id)
            response['info'] = {}
            response['location'] = '/calculation/statushst/{}'.format(id)
            
            
            self.write(dict(response))
            self.redirect("../dashboardhst")

class CalculationStatusHandler(BaseHandler):
    """
    Handlers returning the status of a particular JWST calculation task.
    """
    def get(self, id):
        response = self._get_task_response(id)

        if self.request.connection.stream.closed():
            return

        self.write(dict(response))

        
class CalculationStatusHSTHandler(BaseHandler):
    """
    Handlers returning the status of a particular HST calculation task.
    """
    def get(self, id):
        response = self._get_task_response_hst(id)

        if self.request.connection.stream.closed():
            return

        self.write(dict(response))                

class CalculationDownloadHandler(BaseHandler):
    """
    Handlers returning the downloaded data of a particular calculation task.
    Handlers returning the status of a particular calculation task.
    """
    def get(self, id):
        result = self._get_task_result(id)
  
        if self.request.connection.stream.closed():
            return
        file_name = "ETC-calculation" +id+".p"
 
        with open(os.path.join(__TEMP__,file_name), "wb") as f:
            pickle.dump(result, f)
 
        buf_size = 4096
        self.set_header('Content-Type', 'application/octet-stream')
        self.set_header('Content-Disposition',
                        'attachment; filename=' + file_name)
 
        with open(os.path.join(__TEMP__,file_name), "rb") as f:
            while True:
                data = f.read(buf_size)
                if not data:
                    break
                self.write(data)
        
        allfiles = os.listdir(__TEMP__)
        for i in allfiles:
            if i.find(id) != -1:
                os.remove(os.path.join(__TEMP__,i))
        self.finish()



class CalculationDownloadPandInHandler(BaseHandler):
    """
    Handlers returning the downloaded data of a particular calculation task.
    Handlers returning the status of a particular calculation task.
    """
    def get(self, id):
        result = self._get_task_result(id)
  
        if self.request.connection.stream.closed():
            return
        file_name = "PandExo-Input-file"+id+".txt"
 
        #with open(os.path.join(__TEMP__,file_name), "w") as f:
        #    pickle.dump(result, f)
        np.savetxt(os.path.join(__TEMP__,file_name), np.transpose([result['w'], result['alpha']]))
        
        buf_size = 4096
        self.set_header('Content-Type', 'application/octet-stream')
        self.set_header('Content-Disposition',
                        'attachment; filename=' + file_name)
 
        with open(os.path.join(__TEMP__,file_name), "rb") as f:
            while True:
                data = f.read(buf_size)
                if not data:
                    break
                self.write(data)
        
        allfiles = os.listdir(__TEMP__)
        for i in allfiles:
            if i.find(id) != -1:
                os.remove(os.path.join(__TEMP__,i))


        self.finish()



class CalculationViewHandler(BaseHandler):
    """
    This handler deals with passing the results from Pandeia to the
    `create_component_jwst` function which generates the Bokeh interative plots.
    """
    def get(self, id):
        
        result = self._get_task_result(id)
        
        script, div = create_component_jwst(result)
        div['timing_div'] = result['timing_div']
        div['input_div'] = result['input_div'] 
        div['warnings_div'] = result['warnings_div']

        #delete files
        allfiles = os.listdir(__TEMP__)
        for i in allfiles:
            if i.find(id) != -1:
                os.remove(os.path.join(__TEMP__,i))

        self.render("view.html", script=script, div=div, id=id)



class CalculationViewHSTHandler(BaseHandler):
    """
    This handler deals with passing the results from Pandeia to the
    `create_component_hst` function which generates the Bokeh interative plots.
    """
    def get(self, id):
        result = self._get_task_result(id)
        script, div = create_component_hst(result)
        div['info_div'] = result['info_div']
        self.render("viewhst.html", script=script, div=div, id=id)


def main():
    tornado.options.parse_command_line()
    BaseHandler.executor = ProcessPoolExecutor(max_workers=options.workers)
    http_server = tornado.httpserver.HTTPServer(Application())
    http_server.listen(options.port)
    tornado.ioloop.IOLoop.current().start()


if __name__ == "__main__":
    main()
